from dataclasses import dataclass
from functools import total_ordering

@total_ordering
@dataclass
class Version:
    version: str

    def __post_init__(self):
        self.version_parts = list(map(int, self.version.split('.')))

    def __lt__(self, other):
        if not isinstance(other, Version):
            return NotImplemented
        return self.version_parts < other.version_parts

    def __eq__(self, other):
        if not isinstance(other, Version):
            return NotImplemented
        return self.version_parts == other.version_parts

    def __gt__(self, other):
        if not isinstance(other, Version):
            return NotImplemented
        return self.version_parts > other.version_parts

    def __str__(self):
        return self.version

    def version_string(self):
        return f"v{self.version}"



from typing import List

from frappe_manager.docker_wrapper.subprocess_output import SubprocessOutput


class DockerException(Exception):
    def __init__(
        self,
        command_launched: List[str],
        output: SubprocessOutput,
    ):
        self.docker_command: List[str] = command_launched
        self.output = output

        command_launched_str = " ".join(command_launched)

        error_msg = (
            f"The docker command executed was `{command_launched_str}`.\n"
            f"It returned with code {self.output.exit_code}\n"
        )

        if self.output.stdout:
            error_msg += f"The content of stdout is '{self.output.stdout}'\n"
        else:
            error_msg += "The content of stdout can be found above the " "stacktrace (it wasn't captured).\n"

        if self.output.stderr:
            error_msg += f"The content of stderr is '{self.output.stderr}'\n"
        else:
            error_msg += "The content of stderr can be found above the " "stacktrace (it wasn't captured)."

        super().__init__(error_msg)


class NoSuchContainer(DockerException):
    pass


class NoSuchImage(DockerException):
    pass


class NoSuchService(DockerException):
    pass


class NotASwarmManager(DockerException):
    pass


class NoSuchVolume(DockerException):
    pass




class ServicesComposeNotExist(Exception):
    def __init__(self, message):
        message = message
        super().__init__(message)

class ServicesSecretsDBRootPassNotExist(Exception):
    def __init__(self, message):
        message = message
        super().__init__(message)

class ServicesDBNotStart(Exception):
    def __init__(self, message):
        message = message
        super().__init__(message)

class ServicesException(Exception):
    def __init__(self, message):
        message = message
        super().__init__(message)

class ServicesNotCreated(ServicesException):
    def __init__(self, message: str = 'Not able to generate services compose file.'):
        message = message
        super().__init__(message)

class DatabaseServiceException(Exception):
    def __init__(self, service_name:str , message:str):
        self.message: str = f'Service {service_name} db server : {message}'
        super().__init__(self.message)

class DatabaseServicePasswordNotFound(DatabaseServiceException):
    def __init__(self, service_name: str, message = 'Failed to determine root password.') -> None:
        self.service_name = service_name
        self.message = message.format(self.service_name)
        super().__init__(self.service_name,self.message)

class DatabaseServiceUserRemoveFailError(DatabaseServiceException):
    def __init__(self,username: str, service_name: str, message = 'Failed to remove user {}.') -> None:
        self.service_name = service_name
        self.username = username
        self.message = message.format(self.username)
        super().__init__(self.service_name,self.message)

class DatabaseServiceDBRemoveFailError(DatabaseServiceException):
    def __init__(self,db_name: str, service_name: str, message = 'Failed to remove db {}.') -> None:
        self.service_name = service_name
        self.db_name = db_name
        self.message = message.format(db_name)
        super().__init__(self.service_name,self.message)

class DatabaseServiceDBNotFoundError(DatabaseServiceException):
    def __init__(self, db_name: str, service_name: str, message = 'DB not found {}.') -> None:
        self.service_name = service_name
        self.db_name = db_name
        self.message = message.format(db_name)
        super().__init__(self.service_name,self.message)

class DatabaseServiceStartTimeout(DatabaseServiceException):
    def __init__(self,timeout: int, service_name: str, message = 'DB failed to start, waited for {}s.') -> None:
        self.service_name = service_name
        self.message = message.format(timeout)
        super().__init__(self.service_name,self.message)

class DatabaseServiceDBExportFailed(DatabaseServiceException):
    def __init__(self,service_name: str, db_name:str, message = 'DB export failed for db name {}.') -> None:
        self.service_name = service_name
        self.message = message.format(db_name)
        super().__init__(self.service_name,self.message)

class DatabaseServiceDBImportFailed(DatabaseServiceException):
    def __init__(self,service_name: str, db_dump_path:str, message = 'DB import failed for db dump {}.') -> None:
        self.service_name = service_name
        self.message = message.format(db_dump_path)
        super().__init__(self.service_name,self.message)

class DatabaseServiceDBCreateFailed(DatabaseServiceException):
    def __init__(self,service_name: str, db_name:str, message = 'DB create failed for db name {}.') -> None:
        self.service_name = service_name
        self.message = message.format(db_name)
        super().__init__(self.service_name,self.message)



from typing import Optional
import typer
from typing import List, Optional, Set
from frappe_manager.site_manager.site_exceptions import BenchNotFoundError
from frappe_manager.utils.helpers import check_frappe_app_exists, get_current_fm_version, get_sitename_from_current_path
from frappe_manager.display_manager.DisplayManager import richprint
from frappe_manager import CLI_BENCHES_DIRECTORY, STABLE_APP_BRANCH_MAPPING_LIST, DEFAULT_EXTENSIONS
from frappe_manager.utils.site import validate_sitename


def apps_list_validation_callback(value: List[str] | None):
    """
    Validate the list of apps provided.

    Args:
        value (List[str] | None): The list of apps to validate.

    Raises:
        typer.BadParameter: If the list contains the 'frappe' app, or if any app is invalid or has an invalid branch.

    Returns:
        List[str] | None: The validated list of apps.
    """
    apps_list = []

    if value:
        for app in value:
            appx = app.split(":")

            if appx == "frappe":
                raise typer.BadParameter("'frappe' should not be included here.")

            if "https:" in app or "http:" in app:
                temp_appx = appx
                appx = [":".join(appx[:2])]

                if len(temp_appx) == 3:
                    appx.append(temp_appx[2])

                elif len(temp_appx) > 3:
                    appx.append(temp_appx[2])
                    appx.append(temp_appx[2])

            if len(appx) > 2:
                richprint.stop()
                msg = (
                    "Specify the app in the format <appname>:<branch> or <appname>."
                    "\n<appname> can be a URL or, if it's a FrappeVerse app, simply provide it as 'erpnext' or 'hrms:develop'."
                )
                raise typer.BadParameter(msg)

            if len(appx) == 1:
                exists = check_frappe_app_exists(appx[0])

                if not exists["app"]:
                    richprint.stop()
                    raise typer.BadParameter(f"Invalid app '{appx[0]}'.")

                if appx[0] in STABLE_APP_BRANCH_MAPPING_LIST:
                    appx.append(STABLE_APP_BRANCH_MAPPING_LIST[appx[0]])

            if len(appx) == 2:
                exists = check_frappe_app_exists(appx[0], appx[1])

                if not exists["app"]:
                    richprint.stop()
                    raise typer.BadParameter(f"Invalid app '{appx[0]}'.")

                if not exists["branch"]:
                    richprint.stop()
                    raise typer.BadParameter(f"Invaid branch '{appx[1]}' for '{appx[0]}'.")

            appx = ":".join(appx)
            apps_list.append(appx)
    return apps_list


def frappe_branch_validation_callback(value: str):
    """
    Validate the given Frappe branch.

    Args:
        value (str): The Frappe branch to validate.

    Returns:
        str: The validated Frappe branch.

    Raises:
        typer.BadParameter: If the Frappe branch is not valid.
    """
    if value:
        exists = check_frappe_app_exists("frappe", value)
        if exists["branch"]:
            return value
        else:
            raise typer.BadParameter(f"Frappe branch -> {value} is not valid!! ")


def version_callback(version: Optional[bool] = None):
    """
    Callback function to handle version option.

    Args:
        version (bool, optional): If True, prints the current FM version and exits. Defaults to None.
    """
    if version:
        fm_version = get_current_fm_version()
        richprint.print(fm_version, emoji_code="")
        raise typer.Exit()


def sites_autocompletion_callback():
    sites_list = []
    for dir in CLI_BENCHES_DIRECTORY.iterdir():
        if dir.is_dir():
            dir = dir / "docker-compose.yml"
            if dir.exists() and dir.is_file():
                sites_list.append(sites_list)
    return sites_list


def sitename_callback(sitename: Optional[str]):
    if not sitename:
        sitename = get_sitename_from_current_path()

    if not sitename:
        raise typer.BadParameter(message="Missing Argument")

    sitename = validate_sitename(sitename)

    # check if bench not exists
    bench_path = CLI_BENCHES_DIRECTORY / sitename

    if not bench_path.exists():
        raise BenchNotFoundError(sitename, bench_path)

    return sitename


def code_command_extensions_callback(extensions: List[str]) -> List[str]:
    extx = extensions + DEFAULT_EXTENSIONS
    unique_ext: Set = set(extx)
    unique_ext_list: List[str] = [x for x in unique_ext]
    return unique_ext_list


def create_command_sitename_callback(sitename: str):
    # validate the site
    sitename = validate_sitename(sitename)

    # check if already exists
    bench_path = CLI_BENCHES_DIRECTORY / sitename

    if bench_path.exists():
        richprint.exit(f"The bench '{sitename}' already exists at {bench_path}. Aborting operation.")

    return sitename



from pathlib import Path
from enum import Enum

# TODO configure this using config
# sites_dir = Path().home() / __name__.split(".")[0]
CLI_DIR = Path.home() / "frappe"
CLI_FM_CONFIG_PATH = CLI_DIR / "fm_config.toml"
CLI_SITES_ARCHIVE = CLI_DIR / "archived"
CLI_LOG_DIRECTORY = CLI_DIR / 'logs'
CLI_BENCHES_DIRECTORY = CLI_DIR / 'sites'
CLI_SERVICES_DIRECTORY = CLI_DIR / 'services'

CLI_SERVICES_NGINX_PROXY_DIR = CLI_SERVICES_DIRECTORY / 'nginx-proxy'
CLI_SERVICES_NGINX_PROXY_SSL_DIR = CLI_SERVICES_NGINX_PROXY_DIR / 'ssl'

CLI_BENCH_CONFIG_FILE_NAME = 'bench_config.toml'
SSL_RENEW_BEFORE_DAYS = 30


DEFAULT_EXTENSIONS = [
    "dbaeumer.vscode-eslint",
    "esbenp.prettier-vscode",
    "ms-python.python",
    "ms-python.debugpy",
    "ms-python.flake8",
    "ms-python.black-formatter",
    "visualstudioexptteam.vscodeintellicode",
    "VisualStudioExptTeam.intellicode-api-usage-examples",
]


class SiteServicesEnum(str, Enum):
    frappe = "frappe"
    nginx = "nginx"
    mariadb = "mariadb"
    redis_queue = "redis-queue"
    redis_cache = "redis-cache"
    redis_socketio = "redis-socketio"
    schedule = "schedule"
    socketio = "socketio"


STABLE_APP_BRANCH_MAPPING_LIST = {
    "frappe": 'version-15',
    "erpnext": 'version-15',
    "hrms": 'version-15',
}


class EnableDisableOptionsEnum(str, Enum):
    enable = 'enable'
    disable = 'disable'



from enum import Enum
import os
import tomlkit
from pathlib import Path
from typing import Any, List, Optional
from pydantic import BaseModel, Field, model_validator, validator
from frappe_manager import STABLE_APP_BRANCH_MAPPING_LIST
from frappe_manager.metadata_manager import FMConfigManager, FMLetsencryptConfig
from frappe_manager.ssl_manager import LETSENCRYPT_PREFERRED_CHALLENGE, SUPPORTED_SSL_TYPES
from frappe_manager.ssl_manager.certificate import SSLCertificate
from frappe_manager.ssl_manager.letsencrypt_certificate import LetsencryptSSLCertificate
from frappe_manager.utils.helpers import get_container_name_prefix


class FMBenchEnvType(str, Enum):
    prod = 'prod'
    dev = 'dev'


def ssl_certificate_to_toml_doc(cert: SSLCertificate) -> Optional[tomlkit.TOMLDocument]:
    if cert.ssl_type == SUPPORTED_SSL_TYPES.none:
        return None

    model_dict = cert.model_dump(exclude=cert.toml_exclude, exclude_none=True)
    toml_doc = tomlkit.document()

    for key, value in model_dict.items():
        if isinstance(value, Path):
            toml_doc[key] = str(value.absolute())
        else:
            toml_doc[key] = value
    return toml_doc


class BenchConfig(BaseModel):
    name: str = Field(..., description="The name of the bench")
    developer_mode: bool = Field(..., description="Whether developer mode is enabled")
    admin_tools: bool = Field(..., description="Whether admin tools are enabled")
    environment_type: FMBenchEnvType = Field(..., description="The type of environment")
    ssl: SSLCertificate = Field(..., description="The SSL certificate")

    frappe_branch: str = Field(
        default=STABLE_APP_BRANCH_MAPPING_LIST['frappe'], description="The branch of Frappe to use"
    )
    admin_pass: str = Field('admin', description="The admin password")
    root_path: Path = Field(..., description="The root path")
    mariadb_host: str = Field('global-db', description="The host for MariaDB")
    mariadb_root_pass: str = Field(default='/run/secrets/db_root_password', description="The root password for MariaDB")
    apps_list: List[str] = Field(default=[], description="List of apps")
    userid: int = Field(default_factory=os.getuid, description="The user ID of the current process")
    usergroup: int = Field(default_factory=os.getgid, description="The group ID of the current process")

    @property
    def db_name(self):
        return self.name.replace(".", "-")

    @property
    def container_name_prefix(self):
        return get_container_name_prefix(self.name)

    def export_to_toml(self, path: Path) -> bool:
        ssl_toml_doc: Optional[tomlkit.TOMLDocument] = None

        # check if it's SSLCertificate if not then its
        ssl_toml_doc = ssl_certificate_to_toml_doc(self.ssl)

        exclude = {
            'root_path',
            'mariadb_root_pass',
            'userid',
            'mariadb_host',
            'usergroup',
            'apps_list',
            'frappe_branch',
            'admin_pass',
        }

        if ssl_toml_doc is None:
            exclude.add('ssl')

        # Convert the BenchConfig instance to a dictionary
        bench_dict = self.model_dump(exclude=exclude, exclude_none=True)

        if ssl_toml_doc:
            bench_dict['ssl'] = ssl_toml_doc

        # Serialize the dictionary to a TOML string
        # Create a TOML document from the dictionary
        toml_doc = tomlkit.document()

        for key, value in bench_dict.items():
            if isinstance(value, Path):
                toml_doc[key] = str(value.absolute())
            else:
                toml_doc[key] = value
        try:
            with open(path, 'w') as f:
                f.write(tomlkit.dumps(toml_doc))
            return True
        except Exception as e:
            return False

    @classmethod
    def import_from_toml(cls, path: Path) -> "BenchConfig":
        data = tomlkit.parse(path.read_text())

        data['root_path'] = str(path)

        ssl_data = data.get('ssl', None)

        if ssl_data:
            domain: str = data.get('name', None)  # Set domain from main data if necessary
            ssl_type = ssl_data.get('ssl_type', SUPPORTED_SSL_TYPES.none)

            if ssl_type == SUPPORTED_SSL_TYPES.le:
                email = ssl_data.get('email', None)

                fm_config_manager = FMConfigManager.import_from_toml()

                pref_challenge_data = ssl_data.get("preferred_challenge", None)

                api_token = ssl_data.get('api_token', None)

                if not api_token:
                    api_token = fm_config_manager.letsencrypt.api_token

                api_key = ssl_data.get('api_key', None)

                if not api_key:
                    api_key = fm_config_manager.letsencrypt.api_key

                if not pref_challenge_data:
                    if fm_config_manager.letsencrypt.exists:
                        preferred_challenge = LETSENCRYPT_PREFERRED_CHALLENGE.dns01
                    else:
                        preferred_challenge = LETSENCRYPT_PREFERRED_CHALLENGE.http01
                else:
                    preferred_challenge = pref_challenge_data

                ssl_instance = LetsencryptSSLCertificate(
                    domain=domain,
                    ssl_type=ssl_type,
                    email=email,
                    preferred_challenge=preferred_challenge,
                    api_key=api_key,
                    api_token=api_token,
                )
            else:
                ssl_instance = SSLCertificate(domain=domain, ssl_type=SUPPORTED_SSL_TYPES.none)
        else:
            ssl_instance = SSLCertificate(domain=data.get('name', None), ssl_type=SUPPORTED_SSL_TYPES.none)

        input_data = {
            'name': data.get('name', None),
            'developer_mode': data.get('developer_mode', None),
            'admin_tools': data.get('admin_tools', False),
            'environment_type': data.get('environment_type', None),
            'root_path': data.get('root_path', None),
            'ssl': ssl_instance,
        }

        bench_config_instance = cls(**input_data)

        return bench_config_instance

    def export_to_compose_inputs(self):
        environment = {
            "frappe": {
                "USERID": self.userid,
                "USERGROUP": self.usergroup,
                "APPS_LIST": ",".join(self.apps_list) if self.apps_list else None,
                "FRAPPE_BRANCH": self.frappe_branch,
                "DEVELOPER_MODE": self.developer_mode,
                "ADMIN_PASS": self.admin_pass,
                "DB_NAME": self.db_name,
                "SITENAME": self.name,
                "MARIADB_HOST": self.mariadb_host,
                "MARIADB_ROOT_PASS": self.mariadb_root_pass,
                "CONTAINER_NAME_PREFIX": self.container_name_prefix,
                "ENVIRONMENT": self.environment_type.value,
            },
            "nginx": {
                "SITENAME": self.name,
                "VIRTUAL_HOST": self.name,
                "VIRTUAL_PORT": 80,
                "HSTS": self.ssl.hsts,
            },
            "worker": {
                "USERID": self.userid,
                "USERGROUP": self.usergroup,
            },
            "schedule": {
                "USERID": self.userid,
                "USERGROUP": self.usergroup,
            },
            "socketio": {
                "USERID": self.userid,
                "USERGROUP": self.usergroup,
            },
        }

        users: dict = {"nginx": {"uid": self.userid, "gid": self.usergroup}}
        template_inputs: dict = {
            "environment": environment,
            "user": users,
        }
        return template_inputs



import shutil
from frappe_manager.migration_manager.backup_manager import BackupManager
from frappe_manager.migration_manager.migration_base import MigrationBase
from frappe_manager import CLI_DIR
from frappe_manager.migration_manager.migration_helpers import MigrationBenches
from frappe_manager.migration_manager.version import Version
from frappe_manager.display_manager.DisplayManager import richprint


class MigrationV090(MigrationBase):
    version = Version("0.9.0")

    def init(self):
        self.benches_dir = CLI_DIR / "sites"
        self.backup_manager = BackupManager(str(self.version))

        if self.benches_dir.exists():
            self.skip = True

    def up(self):
        if self.skip:
            return True

        richprint.stdout.rule(f':package: [bold][blue]v{str(self.version)}[/blue][bold]')
        self.logger.info(f"v{str(self.version)}: Started")
        self.logger.info("-" * 40)

        move_directory_list = []

        for site_dir in CLI_DIR.iterdir():
            if site_dir.is_dir():
                docker_compose_path = site_dir / "docker-compose.yml"

                if docker_compose_path.exists():
                    move_directory_list.append(site_dir)

        self.benches_dir.mkdir(parents=True, exist_ok=True)

        benches = MigrationBenches(self.benches_dir)
        benches.stop_benches()

        # move all the directories
        richprint.change_head(f"Moving benches from {CLI_DIR} to {self.benches_dir}")

        for site in move_directory_list:
            site_name = site.parts[-1]
            new_path = self.benches_dir / site_name
            shutil.move(site, new_path)
            self.logger.debug(f"Moved:{site.exists()}")

        richprint.print(f"Moved benches from {CLI_DIR} to {self.benches_dir}")
        self.logger.info("-" * 40)

    def down(self):
        if self.skip:
            return True

        richprint.change_head(f"Working on v{str(self.version)} rollback.")
        self.logger.info("-" * 40)

        if self.benches_dir.exists():
            move_directory_list = []
            for site_dir in self.benches_dir.iterdir():
                if site_dir.is_dir():
                    docker_compose_path = site_dir / "docker-compose.yml"

                    if docker_compose_path.exists():
                        move_directory_list.append(site_dir)

            # stop all the sites
            sites_mananger = MigrationBenches(self.benches_dir)
            sites_mananger.stop_benches()

            # move all the directories
            for site in move_directory_list:
                site_name = site.parts[-1]
                new_path = self.benches_dir.parent / site_name
                shutil.move(site, new_path)

        # delete the sitedir
        shutil.rmtree(self.benches_dir)

        richprint.print(f"[bold]v{str(self.version)}[/bold] rollback successfull.")
        self.logger.info("-" * 40)



import copy
import time
import itertools
from datetime import datetime
import shlex
import shutil
import json
import subprocess
from typing import Any, Dict, List, Optional
from pathlib import Path
from rich.table import Table
from frappe_manager.compose_project.compose_project import ComposeProject
from frappe_manager.docker_wrapper.DockerException import DockerException
from frappe_manager.compose_manager.ComposeFile import ComposeFile
from frappe_manager.display_manager.DisplayManager import richprint
from frappe_manager.logger import log
from frappe_manager.migration_manager.backup_manager import BackupManager
from frappe_manager.services_manager.services import ServicesManager
from frappe_manager.site_manager import VSCODE_LAUNCH_JSON, VSCODE_SETTINGS_JSON, VSCODE_TASKS_JSON
from frappe_manager.site_manager.admin_tools import AdminTools
from frappe_manager.site_manager.bench_config import BenchConfig, FMBenchEnvType
from frappe_manager.site_manager.site_exceptions import (
    BenchAttachTocontainerFailed,
    BenchException,
    BenchFailedToRemoveDevPackages,
    BenchFrappeServiceSupervisorNotRunning,
    BenchNotRunning,
    BenchRemoveDirectoryError,
    BenchSSLCertificateAlreadyIssued,
    BenchSSLCertificateNotIssued,
    BenchServiceNotRunning,
    BenchWorkersSupervisorConfigurtionGenerateError,
)
from frappe_manager.site_manager.workers_manager.SiteWorker import BenchWorkers
from frappe_manager.ssl_manager import LETSENCRYPT_PREFERRED_CHALLENGE, SUPPORTED_SSL_TYPES
from frappe_manager.ssl_manager.certificate import SSLCertificate
from frappe_manager.ssl_manager.nginxproxymanager import NginxProxyManager
from frappe_manager.ssl_manager.ssl_certificate_manager import SSLCertificateManager
from frappe_manager.utils.helpers import (
    capture_and_format_exception,
    format_ssl_certificate_time_remaining,
    get_current_fm_version,
    log_file,
    get_container_name_prefix,
)
from frappe_manager.utils.docker import host_run_cp
from frappe_manager import (
    CLI_BENCH_CONFIG_FILE_NAME,
    CLI_BENCHES_DIRECTORY,
    CLI_DIR,
    EnableDisableOptionsEnum,
    SiteServicesEnum,
)
from frappe_manager.utils.site import domain_level, generate_services_table, get_bench_db_connection_info


class Bench:
    def __init__(
        self,
        path: Path,
        name: str,
        bench_config: BenchConfig,
        compose_project: ComposeProject,
        bench_workers: BenchWorkers,
        services: ServicesManager,
        workers_check: bool = True,
        admin_tools_check: bool = True,
        verbose: bool = False,
    ) -> None:
        self.path = path
        self.name = name
        self.quiet = not verbose
        self.services = services
        self.backup_path = self.path / 'backups'
        self.bench_config: BenchConfig = bench_config
        self.compose_project: ComposeProject = compose_project
        self.workers = bench_workers
        self.logger = log.get_logger()
        self.proxy_manager: NginxProxyManager = NginxProxyManager('nginx', self.compose_project)
        self.admin_tools: AdminTools = AdminTools(self.name, self.path, self.proxy_manager)

        self.certificate_manager = SSLCertificateManager(
            certificate=self.bench_config.ssl,
            webroot_dir=self.proxy_manager.dirs.html.host,
            proxy_manager=services.proxy_manager,
        )

        if workers_check:
            self.ensure_workers_running_if_available()

        if admin_tools_check:
            self.ensure_admin_tools_running_if_available()

    @classmethod
    def get_object(
        cls,
        bench_name: str,
        services: ServicesManager,
        benches_path: Path = CLI_BENCHES_DIRECTORY,
        bench_config_file_name: str = CLI_BENCH_CONFIG_FILE_NAME,
        workers_check: bool = False,
        admin_tools_check: bool = False,
        verbose: bool = False,
    ) -> 'Bench':
        if domain_level(bench_name) == 0:
            bench_name = bench_name + ".localhost"

        bench_path = benches_path / bench_name
        bench_config_path: Path = bench_path / bench_config_file_name

        compose_file_manager = ComposeFile(bench_path / "docker-compose.yml")
        compose_project: ComposeProject = ComposeProject(compose_file_manager, verbose=verbose)
        workers = BenchWorkers(bench_name, bench_path, not verbose)

        bench_config: BenchConfig = BenchConfig.import_from_toml(bench_config_path)

        parms: Dict[str, Any] = {
            'name': bench_name,
            'path': bench_path,
            'bench_config': bench_config,
            'compose_project': compose_project,
            'bench_workers': workers,
            'services': services,
            'workers_check': workers_check,
            'admin_tools_check': admin_tools_check,
        }
        return cls(**parms)

    def sync_bench_config_configuration(self):
        # set developer_mode based on config
        self.common_bench_config_set({'developer_mode': self.bench_config.developer_mode})

        # dev or prod
        self.switch_bench_env()

        # ssl
        certificate_updated = self.update_certificate(self.bench_config.ssl, raise_error=False)
        if certificate_updated:
            richprint.print("Certificate Updated.")

        # admin tools
        if self.bench_config.admin_tools:
            if not self.admin_tools.compose_project.compose_file_manager.compose_path.exists():
                self.sync_admin_tools_compose()
            else:
                self.admin_tools.enable(force_configure=True)
            richprint.print("Enabled Admin-tools.")

        else:
            if not self.admin_tools.compose_project.compose_file_manager.compose_path.exists():
                richprint.print("Admin tools is already disabled.")
            else:
                self.admin_tools.disable()
                richprint.print("Disabled Admin-tools.")

        self.restart_frappe_server()

    def save_bench_config(self):
        richprint.change_head("Saving bench config changes")
        self.bench_config.export_to_toml(self.bench_config.root_path)
        richprint.print("Saved bench config.")

    @property
    def exists(self):
        return self.path.exists()

    @property
    def frappe_container_name_as_hex(self) -> str:
        """
        Returns the hexadecimal representation of the frappe container name.
        """
        container_name = self.compose_project.compose_file_manager.get_container_names()
        return container_name["frappe"].encode().hex()

    def create(self, is_template_bench: bool = False):
        """
        Creates a new bench using the provided template inputs.

        Args:
            template_inputs (dict): A dictionary containing the template inputs.

        Returns:
            None
        """
        try:
            richprint.change_head(f"Creating Bench Directory")
            self.path.mkdir(parents=True, exist_ok=True)

            richprint.change_head(f"Generating bench compose")
            self.generate_compose(self.bench_config.export_to_compose_inputs())
            self.create_compose_dirs()

            if is_template_bench:
                self.remove_attached_secrets()
                global_db_info = self.services.database_manager.database_server_info
                self.sync_bench_common_site_config(global_db_info.host, global_db_info.port)
                self.save_bench_config()
                richprint.print(f"Created template bench: {self.name}", emoji_code=":white_check_mark:")
                return

            richprint.change_head(f"Starting bench services")
            self.compose_project.start_service(force_recreate=True)
            richprint.print(f"Started bench services.")

            richprint.change_head("Creating bench and bench site.")
            self.frappe_logs_till_start()
            richprint.print("Bench and bench site created.")

            richprint.change_head("Configuring bench workers.")
            self.sync_workers_compose(force_recreate=True)
            richprint.change_head("Configuring bench workers.")
            richprint.update_live()

            richprint.change_head(f"Commencing site status check on bench site.")

            # check if bench is created
            if not self.is_bench_created():
                raise Exception("Bench site is inactive or unresponsive.")

            richprint.print("Bench site is active and responding.")

            self.remove_attached_secrets()

            self.logger.info(f"{self.name}: Bench site is active and responding.")

            self.info()

            if not ".localhost" in self.name:
                richprint.print(
                    f"Please note that You will have to add a host entry to your system's hosts file to access the bench locally."
                )

        except Exception as e:
            richprint.stop()

            richprint.error(f"[red][bold]Error Occured : {e}[/bold][/red]")

            exception_traceback_str = capture_and_format_exception()

            logger = log.get_logger()

            logger.error(f"{self.name}: NOT WORKING\n Exception: {exception_traceback_str}")

            log_path = CLI_DIR / "logs" / "fm.log"

            error_message = [
                "There has been some error creating/starting the bench.",
                f":mag: Please check the logs at {log_path}",
            ]

            richprint.error("\n".join(error_message))

            if self.exists:
                remove_status = self.remove_bench(default_choice=False)

                if not remove_status:
                    self.info()

    def common_bench_config_set(self, config: dict):
        """
        Sets the values in the common_site_config.json file.

        Args:
            config (dict): A dictionary containing the key-value pairs to be set in the common_site_config.json file.
        """
        common_bench_config_path = self.path / "workspace/frappe-bench/sites/common_site_config.json"

        if not common_bench_config_path.exists():
            raise BenchException(self.name, message='common_site_config.json not found.')

        common_site_config = {}

        with open(common_bench_config_path, "r") as f:
            common_site_config = json.load(f)

        for key, value in config.items():
            common_site_config[key] = value

        with open(common_bench_config_path, "w") as f:
            json.dump(common_site_config, f)

    def get_common_bench_config(self):
        common_bench_config_path = self.path / "workspace/frappe-bench/sites/common_site_config.json"

        if not common_bench_config_path.exists():
            raise BenchException(self.name, message='common_site_config.json not found.')

        return json.loads(common_bench_config_path.read_text())

    def generate_compose(self, inputs: dict) -> None:
        """
        Generates the compose file for the site based on the given inputs.

        Args:
            inputs (dict): A dictionary containing the inputs for generating the compose file.

        Returns:
            None
        """
        if "environment" in inputs.keys():
            environments: dict = inputs["environment"]
            self.compose_project.compose_file_manager.set_all_envs(environments)

        if "labels" in inputs.keys():
            labels: dict = inputs["labels"]
            self.compose_project.compose_file_manager.set_all_labels(labels)

        if "user" in inputs.keys():
            user: dict = inputs["user"]
            for container_name in user.keys():
                uid = user[container_name]["uid"]
                gid = user[container_name]["gid"]
                self.compose_project.compose_file_manager.set_user(container_name, uid, gid)

        self.compose_project.compose_file_manager.set_network_alias("nginx", "site-network", [self.name])
        self.compose_project.compose_file_manager.set_container_names(get_container_name_prefix(self.name))
        self.compose_project.compose_file_manager.set_secret_file_path(
            "db_root_password", str(self.services.database_manager.database_server_info.secret_path.absolute())
        )

        self.compose_project.compose_file_manager.set_version(get_current_fm_version())
        self.compose_project.compose_file_manager.set_top_networks_name(
            "site-network", get_container_name_prefix(self.name)
        )
        self.compose_project.compose_file_manager.write_to_file()

    def sync_bench_common_site_config(self, services_db_host: str, services_db_port: int):
        """
        Syncs the common site configuration with the global database information and container prefix.

        This function sets the common site configuration data including the socketio port, database host and port,
        and the Redis cache, queue, and socketio URLs.
        """
        container_prefix = get_container_name_prefix(self.name)

        # set common site config
        common_site_config_data = {
            "socketio_port": "80",
            "db_host": services_db_host,
            "db_port": services_db_port,
            "redis_cache": f"redis://{container_prefix}-redis-cache:6379",
            "redis_queue": f"redis://{container_prefix}-redis-queue:6379",
            "redis_socketio": f"redis://{container_prefix}-redis-cache:6379",
        }
        self.common_bench_config_set(common_site_config_data)

    def create_compose_dirs(self) -> bool:
        """
        Creates the necessary directories for the Compose setup.

        Returns:
            bool: True if the directories are created successfully, False otherwise.
        """
        richprint.change_head("Creating required directories")

        frappe_image = self.compose_project.compose_file_manager.yml["services"]["frappe"]["image"]

        workspace_path = self.path / "workspace"
        workspace_path_abs = str(workspace_path.absolute())

        host_run_cp(
            frappe_image,
            source="/workspace",
            destination=workspace_path_abs,
            docker=self.compose_project.docker,
        )

        configs_path = self.path / "configs"
        configs_path.mkdir(parents=True, exist_ok=True)

        # create nginx dirs
        nginx_dir = configs_path / "nginx"
        nginx_dir.mkdir(parents=True, exist_ok=True)

        nginx_poluate_dir = ["conf"]
        nginx_image = self.compose_project.compose_file_manager.yml["services"]["nginx"]["image"]

        for directory in nginx_poluate_dir:
            new_dir = nginx_dir / directory
            if not new_dir.exists():
                new_dir_abs = str(new_dir.absolute())
                host_run_cp(
                    nginx_image,
                    source="/etc/nginx",
                    destination=new_dir_abs,
                    docker=self.compose_project.docker,
                )

        nginx_subdirs = ["logs", "cache", "run", "html"]

        for directory in nginx_subdirs:
            new_dir = nginx_dir / directory
            new_dir.mkdir(parents=True, exist_ok=True)

        richprint.print("Created all required directories.")

        return True

    def start(self, force: bool = False):
        """
        Starts the bench.
        """

        # Should be done in site manager ?
        global_db_info = self.services.database_manager.database_server_info
        self.sync_bench_common_site_config(global_db_info.host, global_db_info.port)

        richprint.change_head("Starting bench services")
        self.admin_tools.remove_nginx_location_config()
        self.compose_project.start_service(force_recreate=force)
        self.sync_bench_config_configuration()
        self.save_bench_config()
        richprint.print("Started bench services.")

        # start workers if exists
        if self.workers.compose_project.compose_file_manager.exists():
            richprint.change_head("Starting bench workers services")
            self.workers.compose_project.start_service(force_recreate=force)
            richprint.print("Started bench workers services.")

        richprint.change_head('Starting frappe server')
        self.frappe_logs_till_start()
        richprint.print('Started frappe server.')

        self.sync_workers_compose()

    def frappe_logs_till_start(self):
        """
        Retrieves and prints the logs of the 'frappe' service until site supervisor starts.

        Args:
            status_msg (str, optional): Custom status message to display. Defaults to None.
        """
        output = self.compose_project.docker.compose.logs(
            services=["frappe"],
            no_log_prefix=True,
            no_color=True,
            follow=True,
            stream=True,
        )

        if self.quiet:
            richprint.live_lines(
                output,
                padding=(0, 0, 0, 2),
                stop_string="INFO supervisord started with pid",
            )
        else:
            for source, line in output:
                if not source == "exit_code":
                    line = line.decode()

                    if "Updating files:".lower() in line.lower():
                        continue
                    if "[==".lower() in line.lower():
                        print(line)
                        continue
                    richprint.stdout.print(line)
                    if "INFO supervisord started with pid".lower() in line.lower():
                        break

    def stop(self) -> bool:
        """
        Stop the site by stopping the containers.

        Returns:
            bool: True if the site is successfully stopped, False otherwise.
        """
        richprint.change_head(f"Stopping bench services")
        self.compose_project.stop_service()
        richprint.print(f"Stopped bench services.")

        if self.workers.compose_project.compose_file_manager.exists():
            richprint.change_head("Starting bench workers services")
            self.workers.compose_project.stop_service()
            richprint.print("Started bench workers services")

        # stop admin_tools if exists
        if self.admin_tools.compose_project.compose_file_manager.exists():
            richprint.change_head("Stopped bench admin tools services")
            self.admin_tools.compose_project.stop_service()
            richprint.print("Stopped bench admin tools services.")

    def remove_containers_and_dirs(self):
        """
        Removes the site by stopping and removing the containers associated with it,
        and deleting the site directory.

        Returns:
            bool: True if the site is successfully removed, False otherwise.
        """
        # TODO handle low level errors like read only, write only, etc.
        if self.compose_project.compose_file_manager.exists():
            richprint.change_head("Removing bench containers.")
            self.compose_project.down_service(remove_ophans=True, volumes=True)
            richprint.print("Removed bench containers.")
        else:
            richprint.warning('Bench compose file not found. Skipping containers removal.')

        if self.workers.compose_project.compose_file_manager.exists():
            richprint.change_head("Removing bench workers containers.")
            self.workers.compose_project.down_service(remove_ophans=True, volumes=True)
            richprint.print("Removed bench workers containers.")
        else:
            richprint.warning('Bench workers compose file not found. Skipping containers removal.')

        if self.admin_tools.compose_project.compose_file_manager.exists():
            richprint.change_head("Removing bench admin tools containers.")
            self.admin_tools.compose_project.down_service(remove_ophans=True, volumes=True)
            richprint.print("Removed bench admin tools containers.")
        else:
            richprint.warning('Bench admin tools compose file not found. Skipping containers removal.')

        richprint.change_head(f"Removing all bench files and directories.")
        try:
            shutil.rmtree(self.path)
        except PermissionError:
            try:
                images = self.compose_project.compose_file_manager.get_all_images()
                if "frappe" in images:
                    frappe_image = images["frappe"]
                    frappe_image = f"{frappe_image['name']}:{frappe_image['tag']}"
                    self.compose_project.docker.run(
                        image=frappe_image,
                        entrypoint="/bin/sh",
                        command="-c 'chown -R frappe:frappe .'",
                        volume=f"{self.path}/workspace:/workspace",
                        stream=False,
                    )
                    shutil.rmtree(self.path)
            except Exception:
                raise BenchRemoveDirectoryError(self.name, self.path)

        richprint.print(f"Removed all bench files and directories.")

    def is_bench_created(self, retry=60, interval=1) -> bool:
        curl_command = 'curl -I --max-time {retry} --connect-timeout {retry} {headers} {url}'
        url = 'http://localhost'
        headers = ''
        if self.bench_config.environment_type == FMBenchEnvType.prod:
            headers = f"-H 'Host: {self.name}'"

        check_command = curl_command.format(retry=retry, headers=headers, url=url)

        for _ in range(retry):
            try:
                # Execute curl command on frappe service
                result = self.compose_project.docker.compose.exec(
                    service="frappe",
                    command=check_command,
                    stream=False,
                )
                for line in result.stdout:
                    if 'HTTP/1.1 200 OK' in line:
                        return True
            except Exception:
                time.sleep(interval)
        return False

    def sync_workers_compose(self, force_recreate: bool = False):
        self.regenerate_workers_supervisor_conf()
        are_workers_not_changed = self.workers.is_expected_worker_same_as_template()
        if are_workers_not_changed:
            richprint.print("Workers configuration remains unchanged.")
            return

        self.workers.generate_compose()
        self.workers.compose_project.start_service(force_recreate=force_recreate)

    def backup_restore_workers_supervisor(self, backup_manager: BackupManager):
        richprint.print("Rolling back to previous workers configuration.")
        for backup in backup_manager.backups:
            backup_manager.restore(backup, force=True)

    def backup_workers_supervisor_conf(self):
        # TODO this can be given to woker class ?
        backup_workers_manager = BackupManager('workers', self.backup_path)

        # TODO use backup manager
        # take backup
        backup_workers_manager.backup(self.workers.supervisor_config_path)

        if self.workers.supervisor_config_path.exists():
            for file_path in self.workers.config_dir.iterdir():
                file_path_abs = str(file_path.absolute())
                if not file_path.is_file():
                    continue
                if file_path_abs.endswith(".fm.supervisor.conf"):
                    from_path = file_path
                    to_path = file_path.parent / f"{file_path.name}.bak"
                    backup_workers_manager.backup(from_path)
        return backup_workers_manager

    def regenerate_workers_supervisor_conf(self):
        richprint.change_head("Regenerating supervisor.conf.")
        self.backup_workers_supervisor_conf()
        # generate the supervisor.conf
        try:
            bench_setup_supervisor_command = (
                "bench setup supervisor --skip-redis --skip-supervisord --yes --user frappe"
            )
            output = self.compose_project.docker.compose.exec(
                service="frappe",
                command=bench_setup_supervisor_command,
                user="frappe",
                workdir="/workspace/frappe-bench",
                stream=True,
            )
            richprint.live_lines(output, padding=(0, 0, 0, 2))
            generate_split_config_command = "/scripts/divide-supervisor-conf.py config/supervisor.conf"
            output = self.compose_project.docker.compose.exec(
                service="frappe",
                command=generate_split_config_command,
                user="frappe",
                workdir="/workspace/frappe-bench",
                stream=True,
            )
            richprint.live_lines(output, padding=(0, 0, 0, 2))
            return True
        except DockerException:
            richprint.print("Rolling back to previous workers configuration.")
            raise BenchWorkersSupervisorConfigurtionGenerateError(self.name)

    def get_bench_installed_apps_list(self):
        apps_json_file = self.path / "workspace" / "frappe-bench" / "sites" / "apps.json"
        apps_data: dict = {}
        if not apps_json_file.exists():
            return {}
        with open(apps_json_file, "r") as f:
            apps_data = json.load(f)
        return apps_data

    # this can be plugable
    def get_db_connection_info(self):
        return get_bench_db_connection_info(self.name, self.path)

    def remove_attached_secrets(self):
        richprint.change_head(f"Removing bench attached secrets")
        running = False

        if self.compose_project.running:
            running = True
            self.compose_project.stop_service(services=['frappe'])
            if self.workers.compose_project.compose_file_manager.exists():
                self.workers.compose_project.stop_service()

        self.compose_project.compose_file_manager.remove_secrets_from_container("frappe")
        self.compose_project.compose_file_manager.remove_root_secrets_compose()
        self.compose_project.compose_file_manager.write_to_file()

        if running:
            self.start()
            richprint.change_head("Waiting till bench server has created.")
            self.frappe_logs_till_start()
            richprint.print("Bench server has created.")

        richprint.print(f"Removed bench attached secrets.")

    def create_certificate(self):
        self.certificate_manager.generate_certificate()
        self.save_bench_config()

    def has_certificate(self):
        return self.certificate_manager.has_certificate()

    def remove_certificate(self):
        self.certificate_manager.remove_certificate()
        self.bench_config.ssl = SSLCertificate(domain=self.name, ssl_type=SUPPORTED_SSL_TYPES.none)
        self.save_bench_config()

    def update_certificate(self, certificate: SSLCertificate, raise_error: bool = True):
        if certificate.ssl_type == SUPPORTED_SSL_TYPES.le:
            if self.has_certificate():
                if raise_error:
                    raise BenchSSLCertificateAlreadyIssued(self.name)
            else:
                self.certificate_manager.set_certificate(certificate)
                self.bench_config.ssl = certificate
                self.create_certificate()

        elif certificate.ssl_type == SUPPORTED_SSL_TYPES.none:
            if self.has_certificate():
                self.remove_certificate()
            else:
                if not raise_error:
                    return
                raise BenchSSLCertificateNotIssued(self.name)

        return True

    def renew_certificate(self):
        if not self.has_certificate():
            raise BenchSSLCertificateNotIssued(self.name)

        if not self.compose_project.is_service_running('nginx'):
            raise BenchServiceNotRunning(self.name, 'nginx')

        self.certificate_manager.renew_certificate()

    def info(self):
        """
        Retrieves and displays information about the bench.

        This method retrieves various information about the site, such as site URL, site root, database details,
        Frappe username and password, root database user and password, and more. It then formats and displays
        this information using the richprint library.
        """

        richprint.change_head(f"Getting bench info")
        bench_db_info = self.get_db_connection_info()

        db_user = bench_db_info["name"]
        db_pass = bench_db_info["password"]

        services_db_info = self.services.database_manager.database_server_info
        bench_info_table = Table(show_lines=True, show_header=False, highlight=True)

        protocol = 'https' if self.has_certificate() else 'http'

        ssl_service_type = f'{self.bench_config.ssl.ssl_type.value}'

        if self.bench_config.ssl.ssl_type == SUPPORTED_SSL_TYPES.le:
            ssl_service_type = (
                f'[{self.bench_config.ssl.preferred_challenge.value}] {self.bench_config.ssl.ssl_type.value}'
            )

        data = {
            "Bench Url": f"{protocol}://{self.name}",
            "Bench Root": f"[link=file://{self.path.absolute()}]{self.path.absolute()}[/link]",
            "Frappe Username": "administrator",
            "Frappe Password": self.bench_config.admin_pass,
            "Root DB User": services_db_info.user,
            "Root DB Password": services_db_info.password,
            "Root DB Host": services_db_info.host,
            "DB Name": db_user,
            "DB User": db_user,
            "DB Password": db_pass,
            "Environment": self.bench_config.environment_type.value,
            "HTTPS": f'{ssl_service_type.upper()} ({format_ssl_certificate_time_remaining(self.certificate_manager.get_certficate_expiry())})'
            if self.has_certificate()
            else 'Not Enabled',
        }

        if not self.bench_config.admin_tools:
            data['Admin Tools'] = 'Not Enabled'
        else:
            admin_tools_Table = Table(show_lines=True, show_edge=False, pad_edge=False, expand=True)
            admin_tools_Table.add_column("Tool")
            admin_tools_Table.add_column("URL")
            admin_tools_Table.add_row("Mailhog", f"{protocol}://{self.name}/mailhog")
            admin_tools_Table.add_row("Adminer", f"{protocol}://{self.name}/adminer")
            data['Admin Tools'] = admin_tools_Table

        bench_info_table.add_column(no_wrap=True)
        bench_info_table.add_column(no_wrap=True)

        for key in data.keys():
            bench_info_table.add_row(key, data[key])

        # get bench apps data
        apps_json = self.get_bench_installed_apps_list()

        if apps_json:
            bench_apps_list_table = Table(show_lines=True, show_edge=False, pad_edge=False, expand=True)

            bench_apps_list_table.add_column("App")
            bench_apps_list_table.add_column("Version")

            for app in apps_json.keys():
                bench_apps_list_table.add_row(app, apps_json[app]["version"])

            bench_info_table.add_row("Bench Apps", bench_apps_list_table)

        running_bench_services = self.compose_project.get_services_running_status()
        running_bench_workers = self.workers.compose_project.get_services_running_status()
        running_bench_admin_tools = self.admin_tools.compose_project.get_services_running_status()

        if running_bench_services:
            bench_services_table = generate_services_table(running_bench_services)
            bench_info_table.add_row("Bench Services", bench_services_table)

        if running_bench_workers:
            bench_workers_table = generate_services_table(running_bench_workers)
            bench_info_table.add_row("Bench Workers", bench_workers_table)

        if running_bench_admin_tools:
            bench_admin_table = generate_services_table(running_bench_admin_tools)
            bench_info_table.add_row("Bench Admin Tools", bench_admin_table)

        richprint.stdout.print(bench_info_table)

    def shell(self, compose_service: str, user: str | None):
        """
        Spawns a shell for the specified service and user.

        Args:
            service (str): The name of the service.
            user (str | None): The name of the user. If None, defaults to "frappe".

        """
        richprint.change_head(f"Spawning shell")

        if compose_service == "frappe" and not user:
            user = "frappe"

        if not self.compose_project.is_service_running(compose_service):
            richprint.exit(
                f"Cannot spawn shell. [blue]{self.name}[/blue]'s compose service '{compose_service}' not running!"
            )

        richprint.stop()

        non_bash_supported = ["redis-cache", "redis-socketio", "redis-queue"]

        shell_path = "/bin/bash" if compose_service not in non_bash_supported else "sh"

        exec_args: Dict[str, Any] = {"service": compose_service, "command": shell_path}

        if compose_service == "frappe":
            exec_args["command"] = "/usr/bin/zsh"
            exec_args["workdir"] = "/workspace/frappe-bench"

        if user:
            exec_args["user"] = user

        exec_args["capture_output"] = False

        try:
            self.compose_project.docker.compose.exec(**exec_args)
        except DockerException as e:
            richprint.warning(f"Shell exited with error code: {e.output.exit_code}")

    def get_log_file_paths(self):
        base_log_dir = self.path / "workspace" / "frappe-bench" / "logs"
        if self.bench_config.environment_type == FMBenchEnvType.dev:
            bench_dev_server_log_path = base_log_dir / "web.dev.log"
            return [bench_dev_server_log_path]
        else:
            bench_prod_server_log_path_stdout = base_log_dir / "web.log"
            bench_prod_server_log_path_stderr = base_log_dir / "web.error.log"
            return [bench_prod_server_log_path_stderr, bench_prod_server_log_path_stdout]

    def handle_frappe_server_file_logs(self, follow: bool):
        log_generators = []

        try:
            # Get log file paths
            log_file_paths = self.get_log_file_paths()

            # Check how many log files are available
            num_log_files = len(log_file_paths)

            if num_log_files == 0:
                richprint.print("[yellow]No log files found.[/yellow]")
                return

            # Open log files and create generators
            for path in log_file_paths:
                log_generators.append(log_file(open(path, 'r'), follow=follow))

            if follow:
                while True:
                    try:
                        for line in itertools.chain.from_iterable(log_generators):
                            print(line.strip())
                    except StopIteration:
                        time.sleep(0.1)
            else:
                for lines in itertools.zip_longest(*log_generators, fillvalue=""):
                    for line in lines:
                        if line:
                            print(line.strip())

        finally:
            for logfile in log_generators:
                logfile.close()

    def logs(self, follow: bool, service: Optional[SiteServicesEnum] = None):
        """
        Display logs for the site or a specific service.

        Args:
            follow (bool): Whether to continuously follow the logs or not.
            service (str, optional): The name of the service to display logs for. If not provided, logs for the entire site will be displayed.
        """
        richprint.change_head(f"Showing logs")
        try:
            if not service:
                self.handle_frappe_server_file_logs(follow=follow)
            else:
                if not self.compose_project.is_service_running(service):
                    richprint.exit(
                        f"Cannot show logs. [blue]{self.name}[/blue]'s compose service '{service}' not running!"
                    )
                self.compose_project.logs(service.value, follow)

        except KeyboardInterrupt:
            richprint.stdout.print("Detected CTRL+C. Exiting..")
            # for log_file in log_files:
            #     log_file.close()

    def attach_to_bench(self, user: str, extensions: List[str], workdir: str, debugger: bool = False):
        """
        Attaches to a running site's container using Visual Studio Code Remote Containers extension.

        Args:
            user (str): The username to be used in the container.
            extensions (List[str]): List of extensions to be installed in the container.
        """

        if not self.compose_project.running:
            raise BenchNotRunning(self.name)

        # check if vscode is installed
        vscode_path = shutil.which("code")

        if not vscode_path:
            # TODO todo this should be exception
            richprint.exit("Visual Studio Code binary i.e 'code' is not accessible via cli.")

        container_hex = self.frappe_container_name_as_hex

        vscode_cmd = shlex.join(
            [
                vscode_path,
                f"--folder-uri=vscode-remote://attached-container+{container_hex}+{workdir}",
            ]
        )

        extensions.sort()

        vscode_config_without_extension = [
            {
                "remoteUser": user,
                "remoteEnv": {"SHELL": "/bin/zsh"},
                "customizations": {
                    "vscode": {
                        "settings": VSCODE_SETTINGS_JSON,
                    }
                },
            }
        ]

        vscode_config_json = copy.deepcopy(vscode_config_without_extension)
        vscode_config_json[0]['customizations']['vscode']["extensions"] = extensions

        labels = {'devcontainer.metadata': json.dumps(vscode_config_json)}

        try:
            labels_previous = self.compose_project.compose_file_manager.get_labels("frappe")[0]
            labels_previous = json.loads(labels_previous["devcontainer.metadata"])
            extensions_previous = copy.deepcopy(labels_previous["customizations"]["vscode"]["extensions"])

        except KeyError:
            extensions_previous = []

        if not extensions_previous == extensions or not user == user:
            richprint.change_head(f"Configuration changed, regenerating label in bench compose")
            self.compose_project.compose_file_manager.set_labels("frappe", labels)
            self.compose_project.compose_file_manager.write_to_file()
            richprint.print(f"Regenerated bench compose.")
            self.compose_project.start_service(['frappe'])

        # sync debugger files
        if debugger:
            richprint.change_head("Sync vscode debugger configuration")
            dot_vscode_dir = self.path / "workspace" / "frappe-bench" / ".vscode"
            tasks_json_path = dot_vscode_dir / "tasks"
            launch_json_path = dot_vscode_dir / "launch"
            setting_json_path = dot_vscode_dir / "settings"

            dot_vscode_config = {
                tasks_json_path: VSCODE_TASKS_JSON,
                launch_json_path: VSCODE_LAUNCH_JSON,
                setting_json_path: VSCODE_SETTINGS_JSON,
            }

            if not dot_vscode_dir.exists():
                dot_vscode_dir.mkdir(exist_ok=True, parents=True)

            for file_path in [launch_json_path, tasks_json_path, setting_json_path]:
                file_name = f"{file_path.name}.json"
                real_file_path = file_path.parent / file_name
                if real_file_path.exists():
                    backup_tasks_path = (
                        file_path.parent / f"{file_path.name}.{datetime.now().strftime('%d-%b-%y--%H-%M-%S')}.json"
                    )
                    shutil.copy2(real_file_path, backup_tasks_path)
                    richprint.print(f"Backup previous '{file_name}' : {backup_tasks_path}")

                with open(real_file_path, "w+") as f:
                    f.write(json.dumps(dot_vscode_config[file_path]))

            # install black in env
            try:
                self.compose_project.docker.compose.exec(
                    service="frappe", command="/workspace/frappe-bench/env/bin/pip install black", stream=True
                )
            except DockerException as e:
                self.logger.error(f"black installation exception: {capture_and_format_exception()}")
                richprint.warning("Not able to install black in env.")

            richprint.print("Synced vscode debugger configuration.")

        richprint.change_head("Attaching to Container")
        output = subprocess.run(vscode_cmd, shell=True)

        if output.returncode != 0:
            raise BenchAttachTocontainerFailed(self.name, 'frappe')

        richprint.print(f"Attached to frappe service container.")

    def remove_database_and_user(self):
        """
        This function is used to remove db and user of the site at self.name and path at self.path.
        """

        bench_db_info = self.get_db_connection_info()
        richprint.change_head("Removing bench db and db users")
        if "name" in bench_db_info:
            db_name = bench_db_info["name"]
            db_user = bench_db_info["user"]

            if not self.services.database_manager.check_db_exists(db_name):
                richprint.warning(f"Bench db [blue]{db_name}[/blue] not found. Skipping...")
            else:
                self.services.database_manager.remove_db(db_name)
                richprint.print(f"Removed bench db [blue]{db_name}[/blue].")

            if not self.services.database_manager.check_user_exists(db_user):
                richprint.warning(f"Bench db user [blue]{db_user}[/blue] not found. Skipping...")
            else:
                self.services.database_manager.remove_user(db_user, remove_all_host=True)
                richprint.print(f"Removed bench db users [blue]{db_user}[/blue].")

    def remove_bench(self, default_choice: bool = True):
        """
        Removes the site.
        """

        params: Dict[str, Any] = {}
        params['prompt'] = f"🤔 Do you want to remove [bold][green]'{self.name}'[/bold][/green]"
        params['choices'] = ["yes", "no"]

        if default_choice:
            params['default'] = 'no'

        continue_remove = richprint.prompt_ask(**params)

        if continue_remove == "no":
            return False

        richprint.start("Removing bench")

        try:
            self.remove_certificate()
        except Exception as e:
            # self.logger.exception(e)
            richprint.warning(str(e))

        self.remove_database_and_user()
        self.remove_containers_and_dirs()
        return True

    def ensure_workers_running_if_available(self):
        if self.workers.compose_project.compose_file_manager.exists():
            if not self.workers.compose_project.running:
                if self.compose_project.running:
                    self.workers.compose_project.start_service()

    def ensure_admin_tools_running_if_available(self):
        if self.admin_tools.compose_project.compose_file_manager.exists():
            if self.bench_config.admin_tools:
                if not self.admin_tools.compose_project.running:
                    if self.compose_project.running:
                        self.admin_tools.enable()
            else:
                atleast_one_service_running = False

                running_services = self.admin_tools.compose_project.get_services_running_status()
                for service in running_services:
                    if service == 'running':
                        atleast_one_service_running = True

                if atleast_one_service_running:
                    self.admin_tools.disable()

    def sync_admin_tools_compose(self):
        self.admin_tools.generate_compose(self.services.database_manager.database_server_info.host)
        restart_required = self.admin_tools.enable(force_recreate_container=True)
        return restart_required

    def restart_frappe_server(self):
        richprint.change_head("Restarting frappe server")
        restart_command = 'supervisorctl -c /opt/user/supervisord.conf restart all'

        try:
            self.compose_project.docker.compose.exec('frappe', restart_command, stream=False)
        except DockerException as e:
            raise BenchException("frappe", "Faild to restart frappe server.")
        richprint.print("Restarted frappe server.")

    def frappe_service_run_command(self, command: str):
        try:
            self.compose_project.docker.compose.exec('frappe', command, user='frappe', stream=False)
        except DockerException as e:
            raise BenchException("frappe", f"Faild to run {command} in frappe service.")

    def get_apps_dev_requirements(self) -> List[str]:
        """Parse pip requirement string to package name and version"""
        apps_path = self.path / 'workspace' / 'frappe-bench' / 'apps'
        apps_path = apps_path.absolute()

        pattern = '**/pyproject.toml'

        # Find all matching files
        pyproject_files = list(apps_path.glob(pattern))

        import tomlkit

        packages_list = []
        # Print found files
        for pyproject_path in pyproject_files:
            pyproject = tomlkit.parse(pyproject_path.read_text())
            packages = pyproject.get('tool', {}).get('bench', {}).get('dev-dependencies', {})
            for name, version in packages.items():
                full_name = name + version
                packages_list.append(full_name)

        return packages_list

    def remove_dev_packages(self):
        richprint.change_head("Removing dev packages from env.")
        dev_packages = self.get_apps_dev_requirements()
        remove_command = '/workspace/frappe-bench/env/bin/python -m pip uninstall --yes ' + " ".join(dev_packages)
        try:
            self.compose_project.docker.compose.exec('frappe', command=remove_command, stream=False)
        except DockerException as e:
            raise BenchFailedToRemoveDevPackages(self.name)
        richprint.print("Removed dev packages from env.")

    def install_dev_packages(self):
        richprint.change_head("Installing dev packages in env.")
        dev_packages = self.get_apps_dev_requirements()
        install_command = '/workspace/frappe-bench/env/bin/python -m pip install --quiet --upgrade ' + " ".join(
            dev_packages
        )
        try:
            self.compose_project.docker.compose.exec('frappe', command=install_command, stream=False)
        except DockerException as e:
            raise BenchFailedToRemoveDevPackages(self.name)
        richprint.print("Installed dev packages in env.")

    def switch_bench_env(self):
        if not self.is_supervisord_running():
            raise BenchFrappeServiceSupervisorNotRunning(self.name)

        if self.bench_config.environment_type == FMBenchEnvType.dev:
            self.install_dev_packages()
            richprint.change_head(f"Configuring and starting {self.bench_config.environment_type.value} services")
            stop_command = 'supervisorctl -c /opt/user/supervisord.conf stop all'
            self.frappe_service_run_command(stop_command)

            unlink_command = 'rm -rf /opt/user/conf.d/frappe-bench-frappe-web.fm.supervisor.conf'
            self.frappe_service_run_command(unlink_command)

            link_command = 'ln -sfn /opt/user/frappe-dev.conf /opt/user/conf.d/frappe-dev.conf'
            self.frappe_service_run_command(link_command)

            reread_command = 'supervisorctl -c /opt/user/supervisord.conf reread'
            self.frappe_service_run_command(reread_command)

            update_command = 'supervisorctl -c /opt/user/supervisord.conf update'
            self.frappe_service_run_command(update_command)

            start_command = 'supervisorctl -c /opt/user/supervisord.conf start all'
            self.frappe_service_run_command(start_command)
            richprint.print(f"Started {self.bench_config.environment_type.value} services.")

        elif self.bench_config.environment_type == FMBenchEnvType.prod:
            self.remove_dev_packages()
            richprint.change_head(f"Configuring and starting {self.bench_config.environment_type.value} services")
            stop_command = 'supervisorctl -c /opt/user/supervisord.conf stop all'

            self.frappe_service_run_command(stop_command)

            unlink_command = 'rm -rf /opt/user/conf.d/frappe-dev.conf'
            self.frappe_service_run_command(unlink_command)

            link_command = 'ln -sfn /workspace/frappe-bench/config/frappe-bench-frappe-web.fm.supervisor.conf /opt/user/conf.d/frappe-bench-frappe-web.fm.supervisor.conf'
            self.frappe_service_run_command(link_command)

            reread_command = 'supervisorctl -c /opt/user/supervisord.conf reread'
            self.frappe_service_run_command(reread_command)

            update_command = 'supervisorctl -c /opt/user/supervisord.conf update'
            self.frappe_service_run_command(update_command)

            start_command = 'supervisorctl -c /opt/user/supervisord.conf start all'
            self.frappe_service_run_command(start_command)
            richprint.print(f"Started {self.bench_config.environment_type.value} services.")

    def is_supervisord_running(self, interval: int = 2, timeout: int = 30):
        for i in range(timeout):
            try:
                status_command = 'supervisorctl -c /opt/user/supervisord.conf status all'
                output = self.compose_project.docker.compose.exec('frappe', status_command, user='frappe', stream=False)
                return True
            except DockerException as e:
                if any('frappe-bench' in s for s in e.output.combined):
                    return True
                time.sleep(interval)
                continue
        return False



from rich.table import Table
from pathlib import Path
import re
import json

from frappe_manager.compose_manager import DockerVolumeMount, DockerVolumeType
from frappe_manager.display_manager.DisplayManager import richprint
from frappe_manager.site_manager.site_exceptions import BenchException

def generate_services_table(services_status: dict):
    # running site services status
    services_table = Table(
        show_lines=False,
        show_edge=False,
        pad_edge=False,
        show_header=False,
        expand=True,
        box=None,
    )

    services_table.add_column(
        "Service Status", ratio=1, no_wrap=True, width=None, min_width=20
    )
    services_table.add_column(
        "Service Status", ratio=1, no_wrap=True, width=None, min_width=20
    )

    for index in range(0,len(services_status),2):
        first_service_table = None
        second_service_table = None

        try:
            first_service = list(services_status.keys())[index]
            first_service_table = create_service_element(first_service, services_status[first_service])
        except IndexError:
            pass

        try:
            second_service = list(services_status.keys())[index+1]
            second_service_table = create_service_element(second_service, services_status[second_service])
        except IndexError:
            pass

        services_table.add_row(first_service_table, second_service_table)

    return services_table

def create_service_element(service, running_status):
    service_table = Table(
        show_lines=False,
        show_header=False,
        highlight=True,
        expand=True,
        box=None,
    )
    service_table.add_column("Service", justify="left", no_wrap=True)
    service_table.add_column("Status", justify="right", no_wrap=True)
    service_status = "\u2713" if running_status == "running" else "\u2718"
    service_table.add_row(
        f"{service}",
        f"{service_status}",
    )
    return service_table

def parse_docker_volume(volume_string: str, root_volumes:dict, compose_path: Path):

    string_parts = volume_string.split(':')

    if len(string_parts) > 1:

        src = string_parts[0]
        dest = string_parts[0]

        is_bind_mount = True

        if string_parts[0] in root_volumes.keys():
            is_bind_mount = False

        if len(string_parts) > 1:
            dest = string_parts[1]

        volume_type = DockerVolumeType.bind

        if not is_bind_mount:
            volume_type = DockerVolumeType.volume

        docker_volume = DockerVolumeMount(src,dest,volume_type, compose_path)

        return docker_volume

def is_fqdn(hostname: str) -> bool:
    """
    https://en.m.wikipedia.org/wiki/Fully_qualified_domain_name
    """
    if not 1 < len(hostname) < 253:
        return False

    # Remove trailing dot
    if hostname[-1] == '.':
        hostname = hostname[0:-1]

    #  Split hostname into list of DNS labels
    labels = hostname.split('.')

    #  Define pattern of DNS label
    #  Can begin and end with a number or letter only
    #  Can contain hyphens, a-z, A-Z, 0-9
    #  1 - 63 chars allowed
    fqdn = re.compile(r'^[a-z0-9]([a-z-0-9-]{0,61}[a-z0-9])?$', re.IGNORECASE)

    # Check that all labels match that pattern.
    return all(fqdn.match(label) for label in labels)

def is_wildcard_fqdn(hostname: str) -> bool:
    """
    Check if the hostname is a fully qualified domain name (FQDN) with optional wildcard.

    A wildcard domain can be specified with a leading asterisk in the first label (e.g., *.example.com).
    https://en.m.wikipedia.org/wiki/Fully_qualified_domain_name
    """
    if not 1 < len(hostname) < 253:
        return False

    # Remove trailing dot
    if hostname[-1] == '.':
        hostname = hostname[:-1]

    # Split hostname into list of DNS labels
    labels = hostname.split('.')

    # Define pattern for a standard DNS label
    fqdn_pattern = re.compile(r'^[a-z0-9]([a-z0-9-]{0,61}[a-z0-9])?$', re.IGNORECASE)

    # Define pattern for a wildcard DNS label (only valid in the first label)
    wildcard_pattern = re.compile(r'^\*\.?$', re.IGNORECASE)

    status = (wildcard_pattern.match(labels[0])) and all(fqdn_pattern.match(label) for label in labels[1:])

    if status == None:
        status = False

    # Check the first label for wildcard pattern, then check all labels for standard pattern
    return status

def domain_level(domain):
    # Split the domain name into individual parts
    parts = domain.split('.')

    # Return the number of parts minus 1 (excluding the TLD)
    return len(parts) - 1

def validate_sitename(sitename: str) -> str:
    match = is_fqdn(sitename)

    if domain_level(sitename) == 0:
        sitename = sitename + ".localhost"

    if not match:
        richprint.error(f"The {sitename} must follow Fully Qualified Domain Name (FQDN) format.", exception=BenchException(sitename,f"Valid FQDN site name not provided."))

    return sitename

def get_bench_db_connection_info(bench_name: str, bench_path: Path):
    db_info = {}
    site_config_file = bench_path / "workspace" / "frappe-bench" / "sites" / bench_name / "site_config.json"
    if site_config_file.exists():
        with open(site_config_file, "r") as f:
            site_config = json.load(f)
            db_info["name"] = site_config["db_name"]
            db_info["user"] = site_config["db_name"]
            db_info["password"] = site_config["db_password"]
    else:
        db_info["name"] = str(bench_name).replace(".", "-")
        db_info["user"] = str(bench_name).replace(".", "-")
        db_info["password"] = None
    return db_info

def pull_docker_images() -> bool:
    from frappe_manager.compose_manager.ComposeFile import ComposeFile
    from frappe_manager.docker_wrapper.DockerClient import DockerClient
    from frappe_manager.docker_wrapper.DockerException import DockerException

    images_list = []
    docker = DockerClient()
    temp_bench_compose_file_manager = ComposeFile(loadfile=Path('/dev/null/docker-compose.yml'))
    services_manager_compose_file_manager = ComposeFile(
        loadfile=Path('/dev/null/docker-compose.yml'), template_name='docker-compose.services.tmpl'
    )
    admin_tools_manager_compose_file_manager = ComposeFile(
        loadfile=Path('/dev/null/docker-compose.yml'), template_name='docker-compose.admin-tools.tmpl'
    )

    images = temp_bench_compose_file_manager.get_all_images()
    images.update(services_manager_compose_file_manager.get_all_images())
    images.update(admin_tools_manager_compose_file_manager.get_all_images())

    for service, image_info in images.items():
        image = f"{image_info['name']}:{image_info['tag']}"
        images_list.append(image)

    # remove duplicates
    images_list = list(dict.fromkeys(images_list))

    no_error = True
    for image in images_list:
        status = f"[blue]Pulling image[/blue] [bold][yellow]{image}[/yellow][/bold]"
        richprint.change_head(status, style=None)
        try:
            output = docker.pull(container_name=image, stream=True)
            richprint.live_lines(output, padding=(0, 0, 0, 2))
        except DockerException as e:
            no_error = False
            richprint.error(f"[bold][red]Error [/bold][/red]: Failed to pull {image}.")
        richprint.print(f"[green]Pulled[/green] [blue]{image}[/blue].")

    return no_error



import typer
from typing import Annotated, Optional
from frappe_manager import CLI_BENCHES_DIRECTORY
from frappe_manager.site_manager.SiteManager import BenchesManager
from frappe_manager.site_manager.site import Bench
from frappe_manager.site_manager.site_exceptions import BenchSSLCertificateNotIssued
from frappe_manager.ssl_manager.certificate_exceptions import SSLCertificateNotDueForRenewalError
from frappe_manager.utils.callbacks import sitename_callback, sites_autocompletion_callback
from frappe_manager.display_manager.DisplayManager import richprint

ssl_root_command = typer.Typer(no_args_is_help=True, rich_markup_mode="rich")


@ssl_root_command.command()
def delete(
    ctx: typer.Context,
    benchname: Annotated[
        Optional[str],
        typer.Argument(
            help="Name of the bench.", autocompletion=sites_autocompletion_callback, callback=sitename_callback
        ),
    ] = None,
):
    """Delete bench ssl certficate."""

    services_manager = ctx.obj["services"]
    bench = Bench.get_object(benchname, services_manager)
    richprint.change_head("Removing SSL certificate")

    if not bench.has_certificate():
        richprint.exit(f"{benchname} doesn't have SSL certificate issued.")
    bench.remove_certificate()
    richprint.print("Removed SSL certificate.")


@ssl_root_command.command()
def renew(
    ctx: typer.Context,
    benchname: Annotated[
        Optional[str],
        typer.Argument(help="Name of the bench.", autocompletion=sites_autocompletion_callback),
    ] = None,
    all: Annotated[bool, typer.Option(help="Renew ssl cert for all benches.")] = False,
):
    """Renew bench ssl certficate."""

    services_manager = ctx.obj["services"]
    benches = BenchesManager(CLI_BENCHES_DIRECTORY, services=services_manager)

    if all:
        sites_list = benches.get_all_bench().keys()
    else:
        sites_list = [benchname]

    for benchname in sites_list:
        bench = Bench.get_object(benchname, services_manager)
        richprint.change_head("Renew certificate")
        try:
            bench.renew_certificate()
        except (BenchSSLCertificateNotIssued, SSLCertificateNotDueForRenewalError) as e:
            richprint.warning(e.message)

        except Exception as e:
            richprint.warning(str(e))



from datetime import datetime
from frappe_manager import CLI_FM_CONFIG_PATH

from frappe_manager.utils.helpers import format_ssl_certificate_time_remaining


class SSLCertificateNotFoundError(Exception):
    """Exception raised when a certificate is not found."""

    def __init__(self, domain, message="No ssl certificate is issued for {}."):
        self.domain = domain
        self.message = message.format(self.domain)
        super().__init__(self.message)


class SSLCertificateEmailNotFoundError(Exception):
    """Exception raised when a certificate is not found."""

    def __init__(self, domain, message="Please provide email using flag '"):
        self.domain = domain
        self.message = message.format(self.domain)
        super().__init__(self.message)


class SSLDNSChallengeCredentailsNotFound(Exception):
    """Exception raised for dns method required credential not found."""

    def __init__(self, message: str = f"Cloudflare dns credentials not found in {CLI_FM_CONFIG_PATH}"):
        self.message = message
        super().__init__(message)


class SSLCertificateChallengeFailed(Exception):
    """Exception raised when a certificate generation failed."""

    def __init__(
        self,
        challenge: str,
    ):
        self.challenge = challenge
        msg = f'{self.challenge} challenge failed.'
        super().__init__(msg)


class SSLCertificateGenerateFailed(Exception):
    """Exception raised when a certificate generation failed."""

    def __init__(
        self,
    ):
        self.message = f"Certificate generation failed."
        super().__init__(self.message)


class SSLCertificateNotDueForRenewalError(Exception):
    def __init__(
        self,
        domain,
        expiry_date: datetime,
        message='[blue]{}:[/blue] Certificate is not due for renewal will expire in {}.',
    ):
        self.domain = domain
        self.expiry_date = expiry_date
        self.time_remaining_txt = format_ssl_certificate_time_remaining(self.expiry_date)
        self.message = message.format(self.domain, self.time_remaining_txt)
        super().__init__(self.message)



from abc import ABC
from logging import Logger
from pathlib import Path
from frappe_manager import CLI_DIR
from frappe_manager.migration_manager.backup_manager import BackupManager
from frappe_manager.migration_manager.migration_exections import MigrationExceptionInBench
from frappe_manager.migration_manager.migration_helpers import (
    MigrationBench,
    MigrationBenches,
    MigrationServicesManager,
)
from frappe_manager.migration_manager.version import Version
from frappe_manager.logger import log
from frappe_manager.services_manager.database_service_manager import DatabaseServerServiceInfo, MariaDBManager
from frappe_manager.display_manager.DisplayManager import richprint
from frappe_manager.utils.helpers import capture_and_format_exception


# @runtime_checkable
class MigrationBase(ABC):
    version: Version = Version("0.0.0")
    benches_dir: Path = CLI_DIR / "sites"
    skip: bool = False
    migration_executor = None
    logger: Logger = log.get_logger()

    def init(self):
        self.backup_manager = BackupManager(str(self.version), self.benches_dir)
        self.benches_manager = MigrationBenches(self.benches_dir)
        self.services_manager: MigrationServicesManager = MigrationServicesManager(services_path=CLI_DIR / 'services')

    def set_migration_executor(self, migration_executor):
        self.migration_executor = migration_executor

    def get_rollback_version(self):
        return self.version

    def up(self):
        if self.skip:
            return True

        richprint.stdout.rule(f':package: [bold][blue]v{str(self.version)}[/blue][bold]')
        self.logger.info(f"v{str(self.version)}: Started")
        self.logger.info("-" * 40)

        self.init()
        self.services_basic_backup()
        self.migrate_services()
        self.migrate_benches()

        self.logger.info("-" * 40)

    def down(self):
        richprint.change_head(f"Working on v{str(self.version)} rollback.")
        self.logger.info("-" * 40)

        # undo each bench
        for bench_name, bench_data in self.migration_executor.migrate_benches.items():
            if not bench_data['exception']:
                self.undo_bench_migrate(bench_data['object'])

        for backup in self.backup_manager.backups:
            self.backup_manager.restore(backup, force=True)
            # richprint.print(f'Restored {backup.bench}'s {backup.src.name}.')

        self.undo_services_migrate()

        richprint.print(f"[bold]v{str(self.version)}[/bold] rollback successfull.")
        self.logger.info("-" * 40)

    def services_basic_backup(self):
        if not self.services_manager.compose_project.compose_file_manager.exists():
            raise MigrationExceptionInBench(
                f"Services compose at {self.services_manager.compose_project.compose_file_manager} not found."
            )
        self.backup_manager.backup(self.services_manager.compose_project.compose_file_manager.compose_path)

    def migrate_services(self):
        pass

    def undo_services_migrate(self):
        pass

    def migrate_benches(self):
        main_error = False

        # migrate each bench
        for bench_name, bench_path in self.benches_manager.get_all_benches().items():
            bench = MigrationBench(name=bench_name, path=bench_path.parent)

            if bench.name in self.migration_executor.migrate_benches.keys():
                bench_info = self.migration_executor.migrate_benches[bench.name]
                if bench_info['exception']:
                    richprint.print(f"Skipping migration for failed bench [blue]{bench.name}[/blue].")
                    main_error = True
                    continue

            self.migration_executor.set_bench_data(bench, migration_version=self.version)
            try:
                self.bench_basic_backup(bench)
                self.migrate_bench(bench)
            except Exception as e:
                traceback_str = capture_and_format_exception()
                self.logger.error(f"{bench.name} [ EXCEPTION TRACEBACK ]:\n {traceback_str}")
                richprint.update_live()
                main_error = True
                self.migration_executor.set_bench_data(bench, e, self.version)

                # restore all backup files
                for backup in self.backup_manager.backups:
                    if backup.bench == bench.name:
                        self.backup_manager.restore(backup, force=True)

                self.undo_bench_migrate(bench)
                self.logger.info(f'Undo successfull for bench: {bench.name}')
                bench.compose_project.down_service(volumes=False, timeout=5)

        if main_error:
            raise MigrationExceptionInBench('')

    def bench_basic_backup(self, bench: MigrationBench):
        richprint.print(f"Migrating bench [bold][blue]{bench.name}[/blue][/bold]")

        # backup docker compose.yml
        self.backup_manager.backup(bench.path / "docker-compose.yml", bench_name=bench.name)

        # backup common_site_config.json
        bench_common_site_config = bench.path / "workspace" / "frappe-bench" / "sites" / "common_site_config.json"
        self.backup_manager.backup(bench_common_site_config, bench_name=bench.name)

        # backup site_config.json
        bench_site_config = bench.path / "workspace" / "frappe-bench" / "sites" / bench.name / "site_config.json"
        self.backup_manager.backup(bench_site_config, bench_name=bench.name)

        server_db_info: DatabaseServerServiceInfo = DatabaseServerServiceInfo.import_from_compose_file(
            'global-db', self.services_manager.compose_project
        )

        self.bench_db_backup(
            bench=bench,
            server_db_info=server_db_info,
            services_manager=self.services_manager,
            backup_manager=self.backup_manager,
        )

    def migrate_bench(self, bench: MigrationBench):
        pass

    def undo_bench_migrate(self, bench: MigrationBench):
        pass

    def bench_db_backup(
        self,
        bench: MigrationBench,
        server_db_info: DatabaseServerServiceInfo,
        services_manager: MigrationServicesManager,
        backup_manager: BackupManager,
    ):
        bench_db_info = bench.get_db_connection_info()
        bench_db_name = bench_db_info["name"]

        richprint.change_head(f'Commencing db {bench.name} backup')

        mariadb_manager = MariaDBManager(server_db_info, services_manager.compose_project)

        from datetime import datetime

        current_datetime = datetime.now()
        formatted_date = current_datetime.strftime("%d-%m-%Y--%H-%M-%S")

        container_backup_dir: Path = Path("/var/log/mysql")
        host_backup_dir: Path = services_manager.services_path / 'mariadb' / 'logs'
        db_sql_file_name = f"db-{bench.name}-{formatted_date}.sql"

        host_db_sql_file_path: Path = host_backup_dir / db_sql_file_name
        container_db_sql_file_path: Path = container_backup_dir / db_sql_file_name

        backup_gz_file_backup_data_path: Path = (
            bench.path / backup_manager.bench_backup_dir / self.version.version / f'{db_sql_file_name}.gz'
        )

        mariadb_manager.db_export(bench_db_name, container_db_sql_file_path)

        import gzip
        import shutil

        # Compress the file using gzip
        with open(host_db_sql_file_path, 'rb') as f_in:
            with gzip.open(backup_gz_file_backup_data_path, 'wb') as f_out:
                shutil.copyfileobj(f_in, f_out)

        host_db_sql_file_path.unlink()

        richprint.print(f'[blue]{bench.name}[/blue] db backup completed successfully.')



import logging
import logging.handlers
import os
from frappe_manager import CLI_LOG_DIRECTORY
import shutil
import gzip
from typing import Dict, Optional
from frappe_manager.display_manager.DisplayManager import richprint

# Define MESSAGE log level
CLEANUP = 25

def namer(name):
    return name + ".gz"

def rotator(source, dest):
    with open(source, 'rb') as f_in:
        with gzip.open(dest, 'wb') as f_out:
            shutil.copyfileobj(f_in, f_out)
    os.remove(source)

loggers: Dict[str, logging.Logger] = {}

# "Register" new loggin level
logging.addLevelName(CLEANUP, 'CLEANUP')

class FMLOGGER(logging.Logger):
    def cleanup(self, msg, *args, **kwargs):
        if self.isEnabledFor(CLEANUP):
            self._log(CLEANUP, msg, args, **kwargs)

def get_logger(log_dir=CLI_LOG_DIRECTORY, log_file_name='fm') -> logging.Logger:
    """ Creates a Log File and returns Logger object """
    # Build Log File Full Path
    logPath = log_dir / f"{log_file_name}.log"

    try:
        log_dir.mkdir(parents=False, exist_ok=True)
    except PermissionError as e:
        richprint.exit(f"Logging not working. {e}",os_exit=True)

    # Create logger object and set the format for logging and other attributes
    if loggers.get(log_file_name):
        logger: Optional[logging.Logger] = loggers.get(log_file_name)
    else:
        logging.setLoggerClass(FMLOGGER)
        logger: Optional[logging.Logger] = logging.getLogger(log_file_name)
        logger.setLevel(logging.DEBUG)

        # configured to roatate after 10 mb
        handler = logging.handlers.RotatingFileHandler(logPath,'a+',maxBytes=10485760, backupCount=3)
        handler.setFormatter(logging.Formatter('[%(asctime)s] %(levelname)s: %(message)s'))
        handler.rotator = rotator
        logger.addHandler(handler)

        # save logger to dict loggers
        loggers[log_file_name] = logger

    return logger



import shutil
import platform
import os
from jinja2 import Template
import typer
from datetime import datetime
from pathlib import Path
from typing import Optional, Any

from frappe_manager import CLI_DIR, CLI_SERVICES_DIRECTORY
from frappe_manager.compose_project.compose_project import ComposeProject
from frappe_manager.services_manager.database_service_manager import (
    DatabaseServerServiceInfo,
    DatabaseServiceManager,
    MariaDBManager,
)
from frappe_manager.services_manager.services_exceptions import (
    ServicesComposeNotExist,
    ServicesException,
    ServicesNotCreated,
)
from frappe_manager.display_manager.DisplayManager import richprint
from frappe_manager.compose_manager.ComposeFile import ComposeFile
from frappe_manager.ssl_manager.nginxproxymanager import NginxProxyManager
from frappe_manager.utils.helpers import (
    get_current_fm_version,
    get_template_path,
    random_password_generate,
    check_and_display_port_status,
    get_unix_groups,
)
from frappe_manager.utils.docker import host_run_cp
from frappe_manager.docker_wrapper.DockerException import DockerException


class ServicesManager:
    def __init__(
        self,
        path=CLI_SERVICES_DIRECTORY,
        verbose: bool = False,
    ) -> None:
        self.path = path
        self.quiet = not verbose
        self.compose_path = self.path / "docker-compose.yml"
        self.typer_context: Optional[typer.Context] = None

    def entrypoint_checks(self, start=False):
        if not self.path.exists():
            try:
                richprint.print(f"Creating global services [blue]{', '.join(self.compose_project.compose_file_manager.get_services_list())}[/blue].", emoji_code=":construction:")
                self.path.mkdir(parents=True, exist_ok=True)
                self.create(clean_install=True)

            except Exception as e:
                raise ServicesNotCreated(f"Not able to create global services [blue]{', '.join(self.compose_project.compose_file_manager.get_services_list())}[/blue].")

            self.compose_project.pull_images()

            richprint.print(f"Created global services [blue]{', '.join(self.compose_project.compose_file_manager.get_services_list())}[/blue].")

            if start:
                self.compose_project.start_service()

        if not self.compose_path.exists():
            raise ServicesComposeNotExist(
                f"Seems like global services has taken a down. No compose file found at {self.compose_path}."
            )

        if start:
            if not self.typer_context.invoked_subcommand == "service":

                if not self.compose_project.running:
                    self.are_ports_free()
                    richprint.print(f"Started non running global services [blue]{', '.join(self.compose_project.compose_file_manager.get_services_list())}[/blue].")
                    self.compose_project.start_service()


        self.database_manager: DatabaseServiceManager = MariaDBManager(
            DatabaseServerServiceInfo.import_from_compose_file('global-db', self.compose_project), self.compose_project
        )

    def init(self):
        # check if the global services exits if not then create
        # TODO this should be done by factory
        current_system = platform.system()

        compose_file_manager = ComposeFile(self.compose_path, template_name="docker-compose.services.tmpl")

        if current_system == "Darwin":
            compose_file_manager = ComposeFile(self.compose_path, template_name="docker-compose.services.osx.tmpl")

        self.compose_project = ComposeProject(compose_file_manager=compose_file_manager)
        self.proxy_manager: NginxProxyManager = NginxProxyManager('global-nginx-proxy', self.compose_project)

        self.fm_headers_path: Path = self.proxy_manager.dirs.confd.host / 'fm_headers.conf'
        self.set_frappe_headers_conf()

    def set_frappe_headers_conf(self):
        if self.fm_headers_path.parent.exists():
            template_path: Path = get_template_path('fm_headers.conf.tmpl')
            template = Template(template_path.read_text())
            output = template.render(current_version=f'v{get_current_fm_version()}')
            self.fm_headers_path.write_text(output)

    def set_typer_context(self, ctx: typer.Context):
        """
        The function sets the typer context
        """
        self.typer_context = ctx

    def create(self, backup: bool = False, clean_install: bool = True):
        envs = {
            "global-db": {
                "MYSQL_ROOT_PASSWORD_FILE": '/run/secrets/db_root_password',
                "MYSQL_DATABASE": "root",
                "MYSQL_USER": "admin",
                "MYSQL_PASSWORD_FILE": '/run/secrets/db_password',
            }
        }
        current_system = platform.system()
        inputs: dict[str, Any] = {"environment": envs}
        try:
            user = {
                "global-db": {
                    "uid": os.getuid(),
                    "gid": os.getgid(),
                }
            }

            if not current_system == "Darwin":
                user["global-nginx-proxy"] = {
                    "uid": os.getuid(),
                    "gid": get_unix_groups()["docker"],
                }

            inputs["user"] = user
        except KeyError:
            raise ServicesException(
                "docker group not found in system. Please add docker group to the system and current user to the docker group."
            )

        if backup:
            if self.path.exists():
                backup_path: Path = CLI_DIR / "backups"
                backup_path.mkdir(parents=True, exist_ok=True)
                current_time = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
                backup_dir_name = f"services_{current_time}"
                self.path.rename(backup_path / backup_dir_name)

        shutil.rmtree(self.path)

        # create required directories
        dirs_to_create = [
            "mariadb/conf",
            "mariadb/logs",
            "nginx-proxy/dhparam",
            "nginx-proxy/certs",
            "nginx-proxy/confd",
            "nginx-proxy/htpasswd",
            "nginx-proxy/vhostd",
            "nginx-proxy/html",
            "nginx-proxy/logs",
            "nginx-proxy/run",
            "nginx-proxy/ssl",
            "nginx-proxy/cache",
            "secrets",
        ]

        # set secrets in compose
        self.generate_compose(inputs)

        if current_system == "Darwin":
            self.compose_project.compose_file_manager.remove_container_user('global-nginx-proxy')
            self.compose_project.compose_file_manager.remove_container_user('global-db')
        else:
            dirs_to_create.append("mariadb/data")

        # create dirs
        for folder in dirs_to_create:
            temp_dir = self.path / folder
            try:
                temp_dir.mkdir(parents=True, exist_ok=True)
            except Exception as e:
                raise ServicesNotCreated(f"Failed to create global services required dir {temp_dir.absolute()}.")

        # populate secrets for db
        db_password_path = self.path / 'secrets' / 'db_password.txt'
        db_root_password_path = self.path / 'secrets' / 'db_root_password.txt'

        db_password_path.write_text(random_password_generate(password_length=16, symbols=True))
        db_root_password_path.write_text(random_password_generate(password_length=24, symbols=True))

        # populate mariadb config
        mariadb_conf = self.path / "mariadb/conf"
        mariadb_conf = str(mariadb_conf.absolute())
        host_run_cp(
            image="mariadb:10.6",
            source="/etc/mysql/.",
            destination=mariadb_conf,
            docker=self.compose_project.docker,
        )

        self.set_frappe_headers_conf()

        self.compose_project.compose_file_manager.set_secret_file_path('db_password', str(db_password_path.absolute()))
        self.compose_project.compose_file_manager.set_secret_file_path(
            'db_root_password', str(db_root_password_path.absolute())
        )
        self.compose_project.compose_file_manager.write_to_file()

        if clean_install:
            # remove previous contaniners and volumes
            self.compose_project.docker.compose.down(remove_orphans=True, timeout=10, volumes=True, stream=False)

    def exists(self):
        return (self.path / "docker-compose.yml").exists()

    def generate_compose(self, inputs: dict):
        # TODO do something about this function
        try:
            # handle environment
            if "environment" in inputs.keys():
                environments: dict = inputs["environment"]
                self.compose_project.compose_file_manager.set_all_envs(environments)

            # handle lablels
            if "labels" in inputs.keys():
                labels: dict = inputs["labels"]
                self.compose_project.compose_file_manager.set_all_labels(labels)

            # handle user
            if "user" in inputs.keys():
                user: dict = inputs["user"]
                for container_name in user.keys():
                    uid = user[container_name]["uid"]
                    gid = user[container_name]["gid"]
                    self.compose_project.compose_file_manager.set_user(container_name, uid, gid)

        # TODO do something about this exception
        except Exception as e:
            raise ServicesNotCreated(f"Not able to generate global services compose file.")

    def shell(self, container: str, user: str | None = None):
        richprint.stop()
        shell_path = "/bin/bash"
        try:
            if user:
                self.compose_project.docker.compose.exec(container, user=user, command=shell_path, capture_output=False)
            else:
                self.compose_project.docker.compose.exec(container, command=shell_path, capture_output=False)
        except DockerException as e:
            richprint.warning(f"Shell exited with error code: {e.output.exit_code}")

    def remove_itself(self):
        shutil.rmtree(self.path)

    def are_ports_free(self):
        ports = [80, 443]
        richprint.change_head(f"Verifying ports {', '.join(map(str, ports))} availability.")
        docker_used_ports = self.compose_project.get_host_port_binds()
        check_and_display_port_status(ports, exclude=docker_used_ports)
        richprint.print(f"Global services will utilize ports {', '.join(map(str, ports))}.")



from typing import Optional, List, Self
from pydantic import EmailStr, Field, model_validator, validator
from frappe_manager.ssl_manager import LETSENCRYPT_PREFERRED_CHALLENGE
from frappe_manager.ssl_manager.certificate import SSLCertificate
from frappe_manager.ssl_manager.certificate_exceptions import SSLDNSChallengeCredentailsNotFound
from frappe_manager.display_manager.DisplayManager import richprint


class LetsencryptSSLCertificate(SSLCertificate):
    preferred_challenge: LETSENCRYPT_PREFERRED_CHALLENGE
    email: EmailStr = Field(..., description="Email used by certbot.")
    api_token: Optional[str] = Field(None, description="Cloudflare API token used by Certbot.")
    api_key: Optional[str] = Field(None, description="Cloudflare Global API Key used by Certbot.")
    toml_exclude: Optional[set] = {'domain', 'alias_domains', 'toml_exclude'}

    @model_validator(mode="after")
    def validate_credentials(self) -> Self:
        if self.preferred_challenge == LETSENCRYPT_PREFERRED_CHALLENGE.dns01:
            if self.api_key or self.api_token:
                return self
            else:
                raise SSLDNSChallengeCredentailsNotFound()

        return self

    def get_cloudflare_dns_credentials(self) -> str:
        creds: List[str] = []

        if self.api_key:
            richprint.print('Using Cloudflare GLOBAL API KEY')
            creds.append(f'dns_cloudflare_email = {self.email}\n')
            creds.append(f'dns_cloudflare_api_key = {self.api_key}\n')

        if self.api_token:
            richprint.print('Using Cloudflare API Token')
            creds.append(f'dns_cloudflare_api_token = {self.api_token}\n')

        if not creds:
            raise SSLDNSChallengeCredentailsNotFound()

        return "\n".join(creds)



from pathlib import Path
from ruamel.yaml import YAML
from ruamel.yaml.comments import CommentedMap as OrderedDict, CommentedSeq as OrderedList
from typing import Any, List, Optional
from frappe_manager.compose_manager import DockerVolumeMount
from frappe_manager.compose_manager.compose_file_exceptions import ComposeSecretNotFoundError, ComposeServiceNotFound
from frappe_manager.display_manager.DisplayManager import richprint
from frappe_manager.utils.site import parse_docker_volume
from frappe_manager.utils.helpers import get_template_path, represent_null_empty
from frappe_manager.migration_manager.version import Version

yaml = YAML(typ="rt")
yaml.representer.ignore_aliases = lambda *args: True

# Set the default flow style to None to preserve the null representation
yaml.default_flow_style = False
yaml.default_style = None


class ComposeFile:
    yml: dict[Any, Any]

    def __init__(self, loadfile: Path, template_name: str = "docker-compose.tmpl", template_dir: Optional[str] = None):
        self.compose_path: Path = loadfile
        self.template_name = template_name
        self.is_template_loaded = False

        self.template_dir = 'templates'

        if template_dir:
            self.template_dir = template_dir

        # check for if the docker-compose.yml file is present if not then use template provided
        if self.exists():
            with open(self.compose_path, "r") as f:
                self.yml = yaml.load(f)
        else:
            self.yml = self.load_template()
            self.is_template_loaded = True

    def exists(self):
        """
        Check if the compose file exists.

        Returns:
            bool: True if the compose file exists, False otherwise.
        """
        return self.compose_path.exists()

    def get_compose_path(self):
        """
        Returns the path of the compose file.
        """
        return self.compose_path

    def load_template(self):
        """
        Load the template file and return its contents as a YAML object.

        Returns:
            dict: The contents of the template file as a YAML object.
        """
        template_path: Path = get_template_path(self.template_name, self.template_dir)
        with open(template_path, "r") as f:
            yml = yaml.load(f)
            return yml

    def set_container_names(self, prefix):
        """
        Sets the container names for each service in the Compose file.

        Args:
            prefix (str): The prefix to be added to the container names.
        """
        for service in self.yml["services"].keys():
            self.yml["services"][service]["container_name"] = prefix + f"-{service}"

    def get_container_names(self) -> dict:
        """
        Returns a dictionary of container names for each service defined in the Compose file.

        Returns:
            dict: A dictionary where the keys are service names and the values are container names.
        """
        container_names: dict = {}
        if self.exists():
            services = self.get_services_list()
            for service in services:
                container_names[service] = self.yml["services"][service]["container_name"]
        return container_names

    def get_services_list(self) -> list:
        """
        Returns a list of services defined in the Compose file.

        Returns:
            list: A list of service names.
        """
        return list(self.yml["services"].keys())

    def is_services_name_same_as_template(self):
        """
        Checks if the service names in the current Compose file are the same as the template file.

        Returns:
            bool: True if the service names are the same, False otherwise.
        """
        template_yml = self.load_template()
        template_service_name_list = list(template_yml["services"].keys())
        template_service_name_list.sort()
        current_service_name_list = list(self.yml["services"].keys())
        current_service_name_list.sort()
        return current_service_name_list == template_service_name_list

    def set_user(self, service, uid, gid):
        """
        Set the user for a specific service in the Compose file.

        Args:
            service (str): The name of the service.
            uid (str): The user ID.
            gid (str): The group ID.
        """
        try:
            self.yml["services"][service]["user"] = f"{uid}:{gid}"
        except KeyError as e:
            richprint.error("Issue in docker template. Not able to set user.", e)

    def get_user(self, service):
        """
        Get the user associated with the specified service.

        Args:
            service (str): The name of the service.

        Returns:
            str or None: The user associated with the service, or None if not found.
        """
        try:
            user = self.yml[service]["user"]
            uid = user.split(":")[0]
            uid = user.split(":")[1]

        except KeyError:
            return None
        return user

    def set_top_networks_name(self, networks_name, prefix):
        """
        Sets the name of the top-level network in the Compose file.

        Args:
            networks_name (str): The name of the network.
            prefix (str): The prefix to be added to the network name.
        """
        if not self.yml["networks"][networks_name]:
            self.yml["networks"][networks_name] = {"name": prefix + f"-network"}
        else:
            self.yml["networks"][networks_name]["name"] = prefix + f"-network"

    def set_network_alias(self, service_name, network_name, alias: list = []):
        """
        Sets the network alias for a given service in the Compose file.

        Args:
            service_name (str): The name of the service.
            network_name (str): The name of the network.
            alias (list, optional): List of network aliases to be set. Defaults to [].

        Returns:
            bool: True if the network alias is set successfully, False otherwise.
        """
        if alias:
            try:
                all_networks = self.yml["services"][service_name]["networks"]
                if network_name in all_networks:
                    self.yml["services"][service_name]["networks"][network_name] = {"aliases": alias}
                    return True
            except KeyError as e:
                return False
        else:
            return False

    def get_network_alias(self, service_name, network_name) -> list | None:
        """
        Retrieves the network aliases for a given service and network name.

        Args:
            service_name (str): The name of the service.
            network_name (str): The name of the network.

        Returns:
            list | None: A list of network aliases if found, otherwise None.
        """
        try:
            all_networks = self.yml["services"][service_name]["networks"]
            if not network_name in all_networks:
                return None

            aliases = self.yml["services"][service_name]["networks"][network_name]["aliases"]
            return aliases
        except KeyError as e:
            return None

    def get_version(self):
        """
        Get the version of the compose file.

        Returns:
            int: The version of the compose file, or 0 if the version is not specified.
        """
        try:
            compose_version = self.yml["x-version"]
            return Version(compose_version)
        except KeyError:
            return Version("0.0.0")

    def set_version(self, version):
        """
        Sets the version of the Compose file.

        Args:
            version (str): The version to set.

        Returns:
            None
        """
        self.yml["x-version"] = version

    def get_all_users(self):
        """
        Retrieves a dictionary of all users defined in the Compose file.

        Returns:
            dict: A dictionary where the keys are service names and the values are dictionaries
                  containing the user's UID and GID.
        """
        users: dict = {}

        if self.exists():
            services = self.get_services_list()
            for service in services:
                if "user" in self.yml["services"][service]:
                    user_data = self.yml["services"][service]["user"]
                    uid = user_data.split(":")[0]
                    gid = user_data.split(":")[1]
                    users[service] = {"uid": uid, "gid": gid}
        return users

    def set_all_users(self, users: dict):
        """
        Sets the UID and GID for all services in the ComposeFile.

        Args:
            users (dict): A dictionary containing the service names as keys and the UID and GID as values.
        """
        for service in users.keys():
            self.set_user(service, users[service]["uid"], users[service]["gid"])

    def get_all_envs(self):
        """
        Retrieves all the environment variables for each service in the Compose file.

        Returns:
            dict: A dictionary containing the service names as keys and their respective environment variables as values.
        """
        envs = {}
        for service in self.yml["services"].keys():
            try:
                env = self.yml["services"][service]["environment"]
                envs[service] = env
            except KeyError:
                pass
        return envs

    def set_all_envs(self, environments: dict):
        """
        Sets environment variables for all containers in the Compose file.

        Args:
            environments (dict): A dictionary containing container names as keys and environment variables as values.

        """
        for container_name in environments.keys():
            self.set_envs(container_name, environments[container_name], append=True)

    def get_all_labels(self):
        """
        Retrieves all the labels for each service in the Compose file.

        Returns:
            dict: A dictionary containing the service names as keys and their respective labels as values.
        """
        labels = {}
        for service in self.yml["services"].keys():
            try:
                label = self.yml["services"][service]["labels"]
                labels[service] = label
            except KeyError:
                pass
        return labels

    def set_all_labels(self, labels: dict):
        """
        Sets labels for all containers in the ComposeFile.

        Args:
            labels (dict): A dictionary containing container names as keys and labels as values.
        """
        for container_name in labels.keys():
            self.set_labels(container_name, labels[container_name])

    def get_all_extrahosts(self):
        """
        Returns a dictionary of all the extra hosts for each service in the Compose file.

        Returns:
            dict: A dictionary where the keys are the service names and the values are the extra hosts.
        """
        extrahosts = {}
        for service in self.yml["services"].keys():
            try:
                extrahost = self.yml["services"][service]["extra_hosts"]
                extrahosts[service] = extrahost
            except KeyError:
                pass
        return extrahosts

    def set_all_extrahosts(self, extrahosts: dict, skip_not_found: bool = False):
        """
        Sets the extrahosts for all containers in the ComposeFile.

        Args:
            extrahosts (dict): A dictionary containing container names as keys and their corresponding extrahosts as values.
            skip_not_found (bool, optional): If True, skips setting extrahosts for containers that are not found. Defaults to False.
        """
        for container_name in extrahosts.keys():
            self.set_extrahosts(container_name, extrahosts[container_name])

    def set_envs(self, container: str, env: dict, append=False):
        """
        Sets the environment variables for a specific container in the Compose file.

        Args:
            container (str): The name of the container.
            env (dict): A dictionary containing the environment variables to be set.
            append (bool, optional): If True, appends the new environment variables to the existing ones.
        """
        new_env = OrderedDict(env)

        if append and type(env) == dict:
            prev_env = self.get_envs(container)
            if prev_env:
                if not type(prev_env) == OrderedList:
                    env = OrderedDict(env)
                    new_env = prev_env | env

        try:
            self.yml["services"][container]["environment"] = new_env
        except KeyError as e:
            pass

    def get_envs(self, container: str) -> dict:
        """
        Get the environment variables for a specific container.

        Args:
            container (str): The name of the container.

        Returns:
            dict: A dictionary containing the environment variables for the container.
                  Returns None if the container or environment variables are not found.
        """
        try:
            env = self.yml["services"][container]["environment"]
            return env
        except KeyError:
            return None

    def set_labels(self, container: str, labels: dict):
        """
        Sets the labels for a specific container in the Compose file.

        Args:
            container (str): The name of the container.
            labels (dict): A dictionary containing the labels to be set.

        """
        try:
            self.yml["services"][container]["labels"] = labels
        except KeyError as e:
            pass

    def get_labels(self, container: str) -> dict:
        """
        Get the labels of a specific container.

        Args:
            container (str): The name of the container.

        Returns:
            dict: The labels of the container, or None if the container or labels are not found.
        """
        try:
            labels = self.yml["services"][container]["labels"]
            return labels
        except KeyError:
            return None

    def set_extrahosts(self, container: str, extrahosts: list):
        """
        Set the extra hosts for a specific container in the Compose file.

        Args:
            container (str): The name of the container.
            extrahosts (list): A list of extra hosts to be added.

        """
        try:
            self.yml["services"][container]["extra_hosts"] = extrahosts
        except KeyError as e:
            pass

    def get_extrahosts(self, container: str) -> list:
        """
        Get the extra hosts for a specific container.

        Args:
            container (str): The name of the container.

        Returns:
            list: A list of extra hosts for the container, or None if not found.
        """
        try:
            extra_hosts = self.yml["services"][container]["extra_hosts"]
            return extra_hosts
        except KeyError:
            return None

    def write_to_file(self):
        """
        Writes the Docker Compose file to the specified path.
        """
        try:
            # saving the docker compose to the directory
            with open(self.compose_path, "w") as f:
                yaml.dump(self.yml, f, transform=represent_null_empty)
        except Exception as e:
            richprint.error(f"Error in writing compose file.", e)

    def get_all_volumes(self):
        """
        Get all the root volumes.
        """

        try:
            volumes = self.yml["volumes"]
        except KeyError as e:
            return {}

        return volumes

    def get_all_services_volumes(self) -> List[DockerVolumeMount]:
        """
        Get all the volume mounts.
        """
        volumes_list: List[DockerVolumeMount] = []

        services = self.get_services_list()
        for service in services:
            volumes = self.get_service_volumes(service)
            volumes_list += volumes

        return volumes_list

    def get_service_volumes(self, service: str) -> List[DockerVolumeMount]:
        """
        Get specific service volume mounts.
        """
        volumes_set = set()

        try:
            volumes_list = self.yml["services"][service]["volumes"]
            for volume in volumes_list:
                volumes_set.add(volume)
        except KeyError as e:
            raise ComposeServiceNotFound(service_name=service)

        volumes_list = []

        for volume in volumes_set:
            volumes_list.append((parse_docker_volume(volume, self.get_all_volumes(), self.compose_path)))

        return volumes_list

    def set_service_volumes(self, service: str, volumes: List[DockerVolumeMount]) -> None:
        """
        Set specific service volume mounts.
        """
        try:
            # Convert DockerVolumeMount objects to strings
            volumes_list = [str(volume) for volume in volumes]

            # Set the volumes for the service
            self.yml["services"][service]["volumes"] = volumes_list
        except KeyError as e:
            raise ComposeServiceNotFound(service_name=service)

    def set_secret_file_path(self, secret_name, file_path):
        try:
            self.yml["secrets"][secret_name]["file"] = file_path
        except KeyError:
            richprint.warning("Not able to set secrets in compose.")

    def get_secret_file_path(self, secret_name) -> Path:
        try:
            file_path = self.yml["secrets"][secret_name]["file"]
            return Path(file_path)
        except KeyError:
            raise ComposeSecretNotFoundError(secret_name, str(self.compose_path.absolute()))

    def remove_secrets_from_container(self, container):
        try:
            del self.yml["services"][container]["secrets"]
        except KeyError:
            richprint.warning(f"Not able to remove secrets from {container}.")

    def remove_root_secrets_compose(self):
        try:
            del self.yml["secrets"]
        except KeyError:
            richprint.warning(f"root level secrets not present.")

    def remove_container_user(self, container):
        try:
            del self.yml["services"][container]["user"]
        except KeyError:
            richprint.warning(f"user not present.")

    def get_all_images(self):
        """
        Retrieves all the images for each service in the Compose file.

        Returns:
            dict: A dictionary containing the service names as keys and their respective image names and tags as values.
        """
        images = {}
        for service in self.yml["services"].keys():
            try:
                image = self.yml["services"][service]["image"]
                name, tag = image.split(":") if ":" in image else (image, "latest")
                images[service] = {"name": name, "tag": tag}
            except KeyError:
                pass
        return images

    def set_all_images(self, images: dict):
        """
        Sets the image for all services in the ComposeFile.

        Args:
            images (dict): A dictionary containing the service names as keys and the image names and tags as values.
        """
        for service, image_info in images.items():
            image = f'{image_info["name"]}:{image_info["tag"]}'
            if service in self.yml["services"]:
                self.yml["services"][service]["image"] = image



from pathlib import Path
from datetime import timedelta, datetime
from frappe_manager import SSL_RENEW_BEFORE_DAYS
from frappe_manager.ssl_manager import SUPPORTED_SSL_TYPES
from frappe_manager.ssl_manager.letsencrypt_certificate_service import LetsEncryptCertificateService
from frappe_manager.ssl_manager.no_op_certificate_service import NoOpCertificateService
from frappe_manager.ssl_manager.certificate_exceptions import (
    SSLCertificateNotDueForRenewalError,
    SSLCertificateNotFoundError,
)
from frappe_manager.ssl_manager.certificate import SSLCertificate
from frappe_manager.ssl_manager.nginxproxymanager import NginxProxyManager
from frappe_manager.ssl_manager.ssl_certificate_service import SSLCertificateService
from frappe_manager.utils.helpers import (
    create_symlink,
    get_certificate_expiry_date,
)


class SSLCertificateManager:
    service: SSLCertificateService
    proxy_manager: NginxProxyManager

    def __init__(self, certificate: SSLCertificate, webroot_dir: Path, proxy_manager: NginxProxyManager):
        self.certificate = certificate
        self.proxy_manager = proxy_manager
        self.webroot_dir = webroot_dir
        self.service = self.ssl_service_factory()

    def ssl_service_factory(self):
        # initializing ssl service
        if self.certificate.ssl_type == SUPPORTED_SSL_TYPES.le:
            webroot_dir = self.webroot_dir
            certificate_service = LetsEncryptCertificateService(self.proxy_manager.dirs.ssl.host, webroot_dir)
            return certificate_service

        certificate_service = NoOpCertificateService(Path('/dev/null'))
        return certificate_service

    def set_certificate(self, certificate: SSLCertificate):
        self.certificate = certificate
        self.service = self.ssl_service_factory()

    def has_certificate(self):
        try:
            self.get_certificate_paths()
            return True
        except SSLCertificateNotFoundError:
            return False

    def get_certificate_paths(self):
        try:
            privkey_file_name: str = f"{self.certificate.domain}.key"
            fullchain_file_name: str = f"{self.certificate.domain}.crt"

            paths = []
            for file_name in [privkey_file_name, fullchain_file_name]:
                path = self.proxy_manager.dirs.certs.host / file_name
                path = path.readlink()
                to_remove = Path(self.proxy_manager.dirs.ssl.container)
                relative_path = path.relative_to(to_remove)
                path = self.proxy_manager.dirs.ssl.host / relative_path
                paths.append(path)
            return paths
        except FileNotFoundError:
            raise SSLCertificateNotFoundError(self.certificate.domain)

    def renew_certificate(self):
        if self.needs_renewal():
            self.service.renew_certificate(self.certificate)
            self.proxy_manager.restart()
        else:
            raise SSLCertificateNotDueForRenewalError(self.certificate.domain, self.get_certficate_expiry())

    def get_certficate_expiry(self):
        privkey_path, fullchain_path = self.get_certificate_paths()
        return get_certificate_expiry_date(fullchain_path)

    def needs_renewal(self) -> bool:
        expiry_date_with_minimum_renew_days = self.get_certficate_expiry() - timedelta(days=SSL_RENEW_BEFORE_DAYS)
        today_date = datetime.now()
        if expiry_date_with_minimum_renew_days.tzinfo:
            today_date = today_date.replace(tzinfo=expiry_date_with_minimum_renew_days.tzinfo)

        if not expiry_date_with_minimum_renew_days > today_date:
            return True
        return False

    def __create_certificate_to_domain_link(self, privkey_path: Path, fullchain_path: Path):
        if not self.certificate.ssl_type == SUPPORTED_SSL_TYPES.none:
            create_symlink(self.__get_cert_container_privkey_path(privkey_path), self.get_cert_proxy_privkey_path())
            create_symlink(
                self.__get_cert_container_fullchain_path(fullchain_path), self.get_cert_proxy_fullchain_path()
            )

    def remove_certificate_to_domain_link(self):
        host_cert_proxy_privkey_path = self.get_cert_proxy_privkey_path()
        host_cert_proxy_fullchain_path = self.get_cert_proxy_fullchain_path()

        try:
            host_cert_proxy_fullchain_path.unlink()
            host_cert_proxy_privkey_path.unlink()
        except:
            pass

    def remove_certificate(self):
        self.remove_certificate_to_domain_link()
        self.service.remove_certificate(self.certificate)
        self.proxy_manager.restart()

    def generate_certificate(self):
        privkey_path, fullchain_path = self.service.generate_certificate(self.certificate)
        self.__create_certificate_to_domain_link(privkey_path, fullchain_path)
        self.proxy_manager.restart()

    def get_cert_proxy_fullchain_path(self) -> Path:
        return self.proxy_manager.dirs.certs.host / f"{self.certificate.domain}.crt"

    def get_cert_proxy_privkey_path(self) -> Path:
        return self.proxy_manager.dirs.certs.host / f"{self.certificate.domain}.key"

    def __get_cert_container_privkey_path(self, privkey_path: Path) -> Path:
        part_to_remove = Path(self.proxy_manager.dirs.ssl.host / self.certificate.ssl_type.value)
        relative_path = privkey_path.relative_to(part_to_remove)
        path = self.proxy_manager.dirs.ssl.container / self.certificate.ssl_type.value / relative_path
        return path

    def __get_cert_container_fullchain_path(self, fullchain_path: Path) -> Path:
        part_to_remove = Path(self.proxy_manager.dirs.ssl.host / self.certificate.ssl_type.value)
        relative_path = fullchain_path.relative_to(part_to_remove)
        path = self.proxy_manager.dirs.ssl.container / self.certificate.ssl_type.value / relative_path
        return path



from pathlib import Path
from frappe_manager.migration_manager.migration_base import MigrationBase
from frappe_manager.migration_manager.migration_exections import MigrationExceptionInBench
from frappe_manager.migration_manager.migration_helpers import (
    MigrationBenches,
    MigrationServicesManager,
)
from frappe_manager.display_manager.DisplayManager import richprint
from frappe_manager.migration_manager.version import Version
from frappe_manager.migration_manager.backup_manager import BackupManager


class MigrationV0131(MigrationBase):
    version = Version("0.13.1")

    def init(self):
        self.cli_dir: Path = Path.home() / 'frappe'
        self.benches_dir = self.cli_dir / "sites"
        self.backup_manager = BackupManager(str(self.version), self.benches_dir)
        self.benches_manager = MigrationBenches(self.benches_dir)
        self.services_manager: MigrationServicesManager = MigrationServicesManager(
            services_path=self.cli_dir / 'services'
        )

    def migrate_services(self):
        if not self.services_manager.compose_project.compose_file_manager.exists():
            raise MigrationExceptionInBench(
                f"Services compose at {self.services_manager.compose_project.compose_file_manager} not found."
            )

        richprint.change_head("Adding fm header config to nginx-proxy")

        # create file fmheaders.conf
        fm_headers_conf_path = self.services_manager.services_path / 'nginx-proxy' / 'confd' / 'fm_headers.conf'
        add_header = f'add_header X-Powered-By "Frappe-Manager {self.version.version_string()}";'

        fm_headers_conf_path.write_text(add_header)

        if self.services_manager.compose_project.is_service_running('global-nginx-proxy'):
            self.services_manager.compose_project.docker.compose.up(services=['global-nginx-proxy'], stream=False)
            self.services_manager.compose_project.docker.compose.restart(services=['global-nginx-proxy'], stream=False)

        richprint.print("Added fm header config to nginx-proxy.")



import typer
from frappe_manager.sub_commands.update_command import update_app

self_app = typer.Typer(no_args_is_help=True, rich_markup_mode="rich")
self_app.add_typer(update_app, name="update", help="Check for updates, Update images etc.")



from pathlib import Path
from typing import Tuple, runtime_checkable, Protocol, TYPE_CHECKING

if TYPE_CHECKING:
    from frappe_manager.ssl_manager.certificate import SSLCertificate


@runtime_checkable
class SSLCertificateService(Protocol):
    root_dir: Path

    def renew_certificate(self, certificate: 'SSLCertificate') -> bool:
        ...

    def remove_certificate(self, certificate: 'SSLCertificate') -> bool:
        ...

    def generate_certificate(self, certificate: 'SSLCertificate') -> Tuple[Path, Path]:
        ...



from typing import Optional

class ComposeFileException(Exception):
    def __init__(
        self,
        error_msg: str,
        exception: Optional[Exception] = None
    ):
        error_msg = f"{error_msg}"
        if exception:
            error_msg = f"{error_msg}\nException : {exception}"
        super().__init__(error_msg)

class ComposeServiceNotFound(Exception):
    def __init__(self, service_name: str, message: str = 'Compose service not found.') -> None:
        self.msg = service_name + ' ' + message
        super().__init__(self.msg)

class ComposeSecretNotFoundError(Exception):
    """Exception raised when a Docker Compose secret value is not found."""

    def __init__(self, secret_name, compose_file_path: str, message="Docker Compose at {} secret value not found"):
        self.secret_name = secret_name
        self.compose_file_path = compose_file_path
        self.message = f"{message.format(self.compose_file_path)}: {secret_name}"
        super().__init__(self.message)



from dataclasses import dataclass
from typing import List


@dataclass
class SubprocessOutput:
    stdout: List[str]
    stderr: List[str]
    combined: List[str]
    exit_code: int

    @classmethod
    def from_output(cls, output):
        stdout = []
        stderr = []
        combined = []
        exit_code = 0

        for source, line in output:
            line = line.decode()
            if source == 'exit_code':
                exit_code = int(line)
            else:
                combined.append(line)
            if source == 'stdout':
                stdout.append(line)
            if source == 'stderr':
                stderr.append(line)

        data = {'stdout': stdout, 'stderr': stderr, 'combined': combined, 'exit_code': exit_code}
        return cls(**data)



import os
from pathlib import Path
from queue import Queue
from subprocess import PIPE, Popen, run
from threading import Thread
from typing import Dict, Iterable, Tuple, Union, Optional
from frappe_manager.logger import log
from frappe_manager.docker_wrapper.DockerException import DockerException
from frappe_manager.display_manager.DisplayManager import richprint
from frappe_manager.docker_wrapper.subprocess_output import SubprocessOutput

process_opened = []


def reader(pipe, pipe_name, queue):
    """
    Reads lines from a pipe and puts them into a queue.

    Args:
        pipe (file-like object): The pipe to read from.
        pipe_name (str): The name of the pipe.
        queue (Queue): The queue to put the lines into.
    """
    logger = log.get_logger()
    try:
        with pipe:
            for line in iter(pipe.readline, b""):
                queue_line = line.decode().strip('\n')
                logger.debug(queue_line)
                queue.put((pipe_name, str(queue_line).encode()))
    finally:
        queue.put(None)


def stream_stdout_and_stderr(
    full_cmd: list,
    env: Optional[Dict[str, str]] = None,
) -> Iterable[Tuple[str, bytes]]:
    """
    Executes a command in Docker and streams the stdout and stderr outputs.

    Args:
        full_cmd (list): The command to be executed in Docker.
        env (Dict[str, str], optional): Environment variables to be passed to the Docker container. Defaults to None.

    Yields:
        Tuple[str, bytes]: A tuple containing the source ("stdout" or "stderr") and the output line.

    Raises:
        DockerException: If the Docker command returns a non-zero exit code.

    Returns:
        Iterable[Tuple[str, bytes]]: An iterable of tuples containing the source and output line.
    """
    logger = log.get_logger()
    logger.debug('- -' * 10)
    logger.debug(f"DOCKER COMMAND: {' '.join(full_cmd)}")
    if env is None:
        subprocess_env = None
    else:
        subprocess_env = dict(os.environ)
        subprocess_env.update(env)

    full_cmd = list(map(str, full_cmd))
    process = Popen(full_cmd, stdout=PIPE, stderr=PIPE, env=subprocess_env)

    process_opened.append(process.pid)

    q = Queue()

    # we use deamon threads to avoid hanging if the user uses ctrl+c
    th = Thread(target=reader, args=[process.stdout, "stdout", q])
    th.daemon = True
    th.start()
    th = Thread(target=reader, args=[process.stderr, "stderr", q])
    th.daemon = True
    th.start()

    output = []
    for _ in range(2):
        for source, line in iter(q.get, None):
            output.append((source, line))
            yield source, line

    exit_code = process.wait()

    logger.debug(f"RETURN CODE: {exit_code}")
    logger.debug('- -' * 10)

    output.append(('exit_code', str(exit_code).encode()))
    yield ("exit_code", str(exit_code).encode())

    if exit_code != 0:
        raise DockerException(full_cmd, SubprocessOutput.from_output(output))


def run_command_with_exit_code(
    full_cmd: list,
    stream: bool = True,
    capture_output: bool = True,
    env: Optional[Dict[str, str]] = None,
) -> Union[Iterable[Tuple[str, bytes]], SubprocessOutput]:
    """
    Run a command and return the exit code.

    Args:
        full_cmd (list): The command to be executed as a list of strings.
        env (Dict[str, str], optional): Environment variables to be set for the command. Defaults to None.
        stream (bool, optional): Flag indicating whether to stream the command output. Defaults to True.
    """
    if not stream:
        if not capture_output:
            run_output = run(full_cmd)
            exit_code = run_output.returncode
            if exit_code != 0:
                raise DockerException(full_cmd, SubprocessOutput([], [], [], exit_code))
            return

        stream_output: SubprocessOutput = SubprocessOutput.from_output(stream_stdout_and_stderr(full_cmd))
        return stream_output

    output: Iterable[Tuple[str, bytes]] = stream_stdout_and_stderr(full_cmd)
    return output


def parameter_to_option(param: str) -> str:
    """Converts a parameter to an option.

    Args:
        param (str): The parameter to be converted.

    Returns:
        str: The converted option.
    """
    option = "--" + param.replace("_", "-")
    return option


def parameters_to_options(param: dict, exclude: list = []) -> list:
    """
    Convert a dictionary of parameters to a list of options for a command.

    Args:
        param (dict): The dictionary of parameters.
        exclude (list, optional): A list of keys to exclude from the options. Defaults to [].

    Returns:
        list: The list of options for the command.
    """
    # remove the self parameter
    temp_param: dict = dict(param)

    del temp_param["self"]

    for key in exclude:
        del temp_param[key]

    # remove all parameters which are not booleans
    params: list = []

    for key in temp_param.keys():
        value = temp_param[key]
        key = "--" + key.replace("_", "-")
        if type(value) == bool:
            if value:
                params.append(key)
        if type(value) == int:
            params.append(key)
            params.append(value)
        if type(value) == str:
            if value:
                params.append(key)
                params.append(value)
        if type(value) == list:
            if value:
                params.append(key)
                params += value

    return params


def is_current_user_in_group(group_name) -> bool:
    """Check if the current user is in the given group.

    Args:
        group_name (str): The name of the group to check.

    Returns:
        bool: True if the current user is in the group, False otherwise.
    """

    from frappe_manager.display_manager.DisplayManager import richprint

    import platform

    if platform.system() == 'Linux':
        import grp
        import pwd
        import os

        current_user = pwd.getpwuid(os.getuid()).pw_name
        try:
            docker_gid = grp.getgrnam(group_name).gr_gid
            docker_group_members = grp.getgrgid(docker_gid).gr_mem
            if current_user in docker_group_members:
                return True
            else:
                richprint.error(
                    f"Your current user [blue][b] {current_user} [/b][/blue] is not in the 'docker' group. Please add it and restart your terminal."
                )
                return False
        except KeyError:
            richprint.error(
                f"The group '{group_name}' does not exist. Please create it and add your current user [blue][b] {current_user} [/b][/blue] to it."
            )
            return False
    else:
        return True


def generate_random_text(length=50):
    """
    Generate a random text of specified length.

    Parameters:
    length (int): The length of the random text to be generated. Default is 50.

    Returns:
    str: The randomly generated text.
    """
    import random, string

    alphanumeric_chars = string.ascii_letters + string.digits
    return "".join(random.choice(alphanumeric_chars) for _ in range(length))


def host_run_cp(image: str, source: str, destination: str, docker):
    """Copy files from source to destination using Docker.

    Args:
        image (str): The Docker image to run.
        source (str): The source file or directory path.
        destination (str): The destination file or directory path.
        docker: The Docker client object.
        verbose (bool, optional): Whether to display verbose output. Defaults to False.
    """

    source_container_name = generate_random_text(10)
    dest_path = Path(destination)
    richprint.change_head(f"Populating {dest_path.name} directory.")

    failed: Optional[int] = None

    try:
        output = docker.run(
            image=image,
            name=source_container_name,
            detach=True,
            stream=False,
            command="tail -f /dev/null",
        )
    except DockerException as e:
        print(e)
        failed = 0

    if not failed:
        # cp from the container
        try:
            output = docker.cp(
                source=source,
                destination=destination,
                source_container=source_container_name,
                stream=False,
            )
        except DockerException as e:
            print(e)
            failed = 1

    if not failed:
        # rm the container
        try:
            output = docker.rm(container=source_container_name, force=True, stream=False)
        except DockerException as e:
            print(e)
            failed = 2

    # check if the destination file exists
    if failed:
        if failed > 1:
            if dest_path.exists():
                import shutil

                shutil.rmtree(dest_path)
        if failed == 2:
            try:
                output = docker.rm(container=source_container_name, force=True, stream=False)
            except DockerException as e:
                pass
        # TODO introuduce custom exception to handle this type of cases where if the flow is not completed then it should raise exception which is handled by caller and then site creation check is done
        raise Exception(f"Failed to copy files from {source} to {destination}.")

    elif not Path(destination).exists():
        raise Exception(f"{destination} not found.")

    richprint.change_head(f"Populated {dest_path.name} directory.")



from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from frappe_manager import CLI_SERVICES_DIRECTORY
from typing import Union

class DockerVolumeType(str, Enum):
    volume = 'volume'
    bind = 'bind'

class DockerVolumeMount:
    def __init__(self, host: Union[str,Path], container: str, type: str, compose_path: Path):
        self.host = host
        self.type = type
        self.container = Path(container)
        self.compose_path = compose_path

        if type == DockerVolumeType.bind:

            self.host = Path(self.host)
            # only join ./ paths
            if str(host).startswith('./'):
               self.host = compose_path.parent.joinpath(host)

    def __str__(self):
        source = Path(self.host) if not isinstance(self.host, Path) else self.host

        if self.type == 'bind':
            source = str(self.host).replace(str(self.compose_path.parent), '.')

        dest  = str(self.container)
        return f'{source}:{dest}'



import atexit

from frappe_manager.display_manager.DisplayManager import richprint
from frappe_manager.logger import log
from frappe_manager.utils.helpers import capture_and_format_exception, remove_zombie_subprocess_process
from frappe_manager import CLI_LOG_DIRECTORY
from frappe_manager.utils.docker import process_opened
from frappe_manager.commands import app


def cli_entrypoint():
    try:
        app()
    except Exception as e:
        logger = log.get_logger()

        richprint.error(f'[red]Error Occured[/red]  {str(e).strip()}')
        richprint.error(f"More info about error is logged in {CLI_LOG_DIRECTORY/'fm.log'}", emoji_code=':mag:')
        richprint.stop()

        exception_traceback: str = capture_and_format_exception()
        logger.error(f"Exception Occured:  : \n{exception_traceback}")

    finally:
        atexit.register(exit_cleanup)


def exit_cleanup():
    """
    This function is used to perform cleanup at the exit.
    """
    remove_zombie_subprocess_process(process_opened)
    richprint.stop()



import shlex
from io import StringIO
from pathlib import Path
from typing import List, Tuple
from certbot._internal.main import make_or_verify_needed_dirs
from certbot._internal.plugins import disco as plugins_disco
from certbot._internal import cli, storage
from certbot._internal.display import obj as display_obj
from certbot import crypto_util
from certbot.errors import AuthorizationError
from frappe_manager.logger import log
from frappe_manager.ssl_manager import LETSENCRYPT_PREFERRED_CHALLENGE, SUPPORTED_SSL_TYPES
from frappe_manager.ssl_manager.certificate_exceptions import (
    SSLCertificateChallengeFailed,
    SSLCertificateGenerateFailed,
    SSLCertificateNotFoundError,
)
from frappe_manager.ssl_manager.letsencrypt_certificate import LetsencryptSSLCertificate
from frappe_manager.ssl_manager.ssl_certificate_service import SSLCertificateService
from frappe_manager.display_manager.DisplayManager import richprint


class LetsEncryptCertificateService(SSLCertificateService):
    def __init__(
        self,
        ssl_service_dir: Path,
        webroot_dir: Path,
    ):
        self.webroot_dir = webroot_dir
        self.root_dir = ssl_service_dir / SUPPORTED_SSL_TYPES.le.value

        # certbot dirs
        self.config_dir: Path = self.root_dir / "config"
        self.work_dir: Path = self.root_dir / 'work'
        self.logs_dir: Path = self.root_dir / 'logs'
        self.dns_config_dir = self.root_dir / 'dns_configs'

        self.base_command = f"--work-dir {self.work_dir} --config-dir {self.config_dir} --logs-dir {self.logs_dir}"
        self.logger = log.get_logger()
        self.console_output = StringIO()

    def renew_certificate(self, certificate: LetsencryptSSLCertificate):
        renew_certificate_command = (
            self.base_command + f' renew --cert-name {certificate.domain} --non-interactive --force-renewal'
        )
        renew_certificate_config = self._get_le_config(shlex.split(renew_certificate_command), quiet=True)
        make_or_verify_needed_dirs(renew_certificate_config)
        plugins = plugins_disco.PluginsRegistry.find_all()
        renew_certificate_config.func(renew_certificate_config, plugins)

    def _get_le_config(self, cli_args: List[str], quiet: bool = False):
        plugins = plugins_disco.PluginsRegistry.find_all()

        config = cli.prepare_and_parse_args(plugins, cli_args)
        if quiet:
            config.quiet = True
            config.noninteractive_mode = True
            displayer = display_obj.NoninteractiveDisplay(outfile=self.console_output)
            display_obj.set_display(displayer)
        return config

    def get_certificate_paths(self, certificate: LetsencryptSSLCertificate) -> Tuple[Path, Path]:
        config = self._get_le_config(shlex.split(self.base_command), quiet=True)
        for renewal_file in storage.renewal_conf_files(config):
            renewal_candidate = storage.RenewableCert(renewal_file, config)
            crypto_util.verify_renewable_cert(renewal_candidate)

            # parsed_certs.append(renewal_candidate)
            if certificate.domain == renewal_candidate.lineagename:
                privkey_path = Path(renewal_candidate.key_path)
                fullchain_path = Path(renewal_candidate.fullchain_path)
                return privkey_path, fullchain_path
        raise SSLCertificateNotFoundError(certificate.domain)

    def remove_certificate(self, certificate: LetsencryptSSLCertificate):
        richprint.change_head("Removing Letsencrypt certificate")
        remove_certificate_command = self.base_command + f' delete --cert-name {certificate.domain}'
        remove_certificate_config = self._get_le_config(shlex.split(remove_certificate_command), quiet=True)
        plugins = plugins_disco.PluginsRegistry.find_all()
        remove_certificate_config.func(remove_certificate_config, plugins)
        richprint.print("Removed Letsencrypt certificate")

    def generate_certificate(self, certificate: LetsencryptSSLCertificate):
        gen_command: str = self.base_command + f" certonly "

        richprint.print(f"Using Let's Encrypt {certificate.preferred_challenge.value} challenge.")

        dns_config_path = self.dns_config_dir / f'{certificate.domain}.txt'

        if certificate.preferred_challenge == LETSENCRYPT_PREFERRED_CHALLENGE.http01:
            gen_command += f' --webroot -w {self.webroot_dir}'

        elif certificate.preferred_challenge == LETSENCRYPT_PREFERRED_CHALLENGE.dns01:
            self.dns_config_dir.mkdir(parents=True, exist_ok=True)

            api_creds = certificate.get_cloudflare_dns_credentials()
            dns_config_path.write_text(api_creds)
            dns_config_path.chmod(0o600)

            gen_command += f' --dns-cloudflare --dns-cloudflare-credentials {dns_config_path.absolute()}'

        gen_command += f' --keep-until-expiring --expand'
        gen_command += f' --agree-tos -m "{certificate.email}" --no-eff-email'

        all_domains = [f'{certificate.domain}'] + certificate.alias_domains

        # alias domains
        for alias_domain in all_domains:
            gen_command += f" -d '{alias_domain}'"

        try:
            richprint.change_head("Getting Letsencrypt certificate")
            self.logger.debug(f'Certbot command: {gen_command}')
            config = self._get_le_config(shlex.split(gen_command), quiet=True)
            plugins = plugins_disco.PluginsRegistry.find_all()
            config.func(config, plugins)
            output = '\n'.join(line for line in self.console_output.getvalue().split('\n') if not line.startswith('!!'))
            richprint.stdout.print(output)

        except AuthorizationError as e:
            self.logger.exception(e)
            output = '\n'.join(line for line in self.console_output.getvalue().split('\n') if not line.startswith('!!'))
            richprint.stdout.print(output)
            dns_config_path.unlink()
            raise SSLCertificateChallengeFailed(certificate.preferred_challenge)

        except Exception as e:
            self.logger.exception(e)
            output = '\n'.join(line for line in self.console_output.getvalue().split('\n') if not line.startswith('!!'))
            richprint.stdout.print(output)
            dns_config_path.unlink()
            raise SSLCertificateGenerateFailed()

        richprint.print("Acquired Letsencrypt certificate: Done")
        return self.get_certificate_paths(certificate)



from typing import List
from frappe_manager.compose_manager import DockerVolumeMount, DockerVolumeType
from frappe_manager.compose_project.compose_project import ComposeProject
from frappe_manager.display_manager.DisplayManager import richprint
from frappe_manager.utils.helpers import create_class_from_dict


class NginxProxyManager:
    def __init__(
        self,
        service_name: str,
        compose_project: ComposeProject,
    ):
        self.service_name = service_name
        self.compose_project = compose_project
        self.dirs = self._get_docker_volume_dirs()

    def _get_docker_volume_dirs(self):
        all_volumes: List[DockerVolumeMount] = self.compose_project.compose_file_manager.get_service_volumes(
            self.service_name
        )
        dirs = {}
        for volume in all_volumes:
            if volume.type == DockerVolumeType.bind:
                name = str(volume.host.name)
                dirs[name] = volume
        dirs_class = create_class_from_dict('dirs', dirs)
        return dirs_class()

    def reload(self):
        richprint.change_head("Reloading nginx")

        if self.compose_project.running:
            output = self.compose_project.docker.compose.exec(
                service=self.service_name, command='nginx -s reload', stream=False
            )
            richprint.print("Reloaded nginx.")

    def restart(self):
        richprint.change_head("Restarting nginx")

        if self.compose_project.running:
            output = self.compose_project.docker.compose.restart(services=[self.service_name], stream=False)
            richprint.print("Restarting nginx.")



from pathlib import Path

from pydantic import config


class BenchException(Exception):
    def __init__(
        self,
        bench_name: str,
        message: str,
    ):
        self.message = f"[blue][bold]{bench_name} :[/bold][/blue] {message}"
        super().__init__(self.message)


class BenchDockerComposeFileNotFound(BenchException):
    def __init__(
        self,
        bench_name: str,
        path: Path,
        message: str = 'Compose file not found at {}. Aborting operation.',
    ):
        self.bench_name = bench_name
        self.path = path
        self.message = message.format(self.path)
        super().__init__(self.bench_name, self.message)


class BenchServiceNotRunning(BenchException):
    def __init__(
        self,
        bench_name: str,
        service: str,
        message: str = 'Service {} not running.',
    ):
        self.bench_name = bench_name
        self.service = service
        self.message = message.format(self.service)
        super().__init__(self.bench_name, self.message)


class BenchNotFoundError(BenchException):
    def __init__(
        self,
        bench_name: str,
        path: Path,
        message: str = 'Bench not found at {}.',
    ):
        self.bench_name = bench_name
        self.path = path
        self.message = message.format(self.path)
        super().__init__(self.bench_name, self.message)


class BenchRemoveDirectoryError(BenchException):
    def __init__(
        self,
        bench_name: str,
        path: Path,
        message: str = 'Remove dirs failed at {}.',
    ):
        self.bench_name = bench_name
        self.path = path
        self.message = message.format(self.path)
        super().__init__(self.bench_name, self.message)


class BenchLogFileNotFoundError(BenchException):
    def __init__(
        self,
        bench_name: str,
        path: Path,
        message: str = 'Log file not found at {}.',
    ):
        self.bench_name = bench_name
        self.path = path
        self.message = message.format(self.path)
        super().__init__(self.bench_name, self.message)


class BenchWorkersStartError(BenchException):
    def __init__(
        self,
        bench_name: str,
        message: str = 'Workers not able to start.',
    ):
        self.bench_name = bench_name
        self.message = message
        super().__init__(self.bench_name, self.message)


class BenchWorkersSupervisorConfigurtionGenerateError(BenchException):
    def __init__(
        self,
        bench_name: str,
        message: str = 'Failed to configure workers.',
    ):
        self.bench_name = bench_name
        self.message = message
        super().__init__(self.bench_name, self.message)


class BenchWorkersSupervisorConfigurtionNotFoundError(BenchException):
    def __init__(
        self,
        bench_name: str,
        config_dir: str,
        message: str = 'Superviosrd workers configuration not found in {}.',
    ):
        self.bench_name = bench_name
        self.config_dir = config_dir
        self.message = message.format(self.config_dir)
        super().__init__(self.bench_name, self.message)


class BenchConfigFileNotFound(BenchException):
    def __init__(self, bench_name, config_path, message='Config file not found at {}.'):
        self.bench_name = bench_name
        self.config_path = config_path
        self.message = message.format(config_path)
        super().__init__(self.bench_name, self.message)


class BenchConfigValidationError(BenchException):
    def __init__(self, bench_name, config_path, message='FM bench config not valid at {}'):
        self.bench_name = bench_name
        self.conig_path = config_path
        self.message = message.format(self.conig_path)
        super().__init__(self.bench_name, self.message)


class AdminToolsFailedToStart(BenchException):
    def __init__(self, bench_name, message="Failed to start admin tools."):
        self.bench_name = bench_name
        self.message = message
        super().__init__(self.bench_name, self.message)


class BenchSSLCertificateAlreadyIssued(BenchException):
    def __init__(self, bench_name, message="SSL Certificate already issued."):
        self.bench_name = bench_name
        self.message = message
        super().__init__(self.bench_name, self.message)


class BenchSSLCertificateNotIssued(BenchException):
    def __init__(self, bench_name, message="No SSL Certificate issued."):
        self.bench_name = bench_name
        self.message = message
        super().__init__(self.bench_name, self.message)


class BenchAttachTocontainerFailed(BenchException):
    def __init__(self, bench_name, service_name, message="Attach to {} service container failed."):
        self.bench_name = bench_name
        self.service_name = service_name
        self.message = message.format(self.service_name)
        super().__init__(self.bench_name, self.message)


class BenchNotRunning(BenchException):
    def __init__(self, bench_name, message="Bench services not running."):
        self.bench_name = bench_name
        self.message = message
        super().__init__(self.bench_name, self.message)


class BenchFailedToRemoveDevPackages(BenchException):
    def __init__(self, bench_name, message="Not able pip uninstall dev packages."):
        self.bench_name = bench_name
        self.message = message
        super().__init__(self.bench_name, self.message)


class BenchFrappeServiceSupervisorNotRunning(BenchException):
    def __init__(self, bench_name, message="Supervisorctl is not running in frappe service"):
        self.bench_name = bench_name
        self.message = message
        super().__init__(self.bench_name, self.message)
